c             SPECTRAL CODE NSTAB FOR MHD EQUILIBRIUM AND STABILITY 
c
c     12/29/07 Revised for DELL box224 to run nem14h ...
c     02/20/03 Revised to check "fast" compiling & switch cos vs sin !
c     11/07/01 Revised to optimize MHH2 and fix del0arr!
c     07/28/01 Edited to remove redundant statements; check this works!
c     08/02/01 Fixed real*4 bug and did major editing
c     07/10/01 Dimension fixes for NCAR; edited for SUN and NERSC together
c
c     Query '!' for important changes in the code!
c
c     07/13/01 snstab.f uses sgemm, starting at line 4175 (for the CRAY!)
c              subroutine second has been commented out (for the CRAY!).
c     02/10/00 Listing of code tnstab.f with comment cards.
c              Input file nstabh and output file nstabout are declared
c	       with the open statements for units 5 and 7, respectively.
c     02/03/00 To run on a CRAY, replace the ten calls to dgemm by calls
c              to sgemm, starting on line 4176. Also, comment the
c              subroutine SECOND.
c              Check the iota formula on line 1049.
c              If you do not have BLAS in your machine, compile
c              tnstab.f with the file dblas3.f from netlib.
c     12/08/97 BOOTSTRAP IOTA REVISION ON LINES 677, 2459 AND 3180
c     10/26/98 see tfree.f for how tnstab.f is coupled to tran.
c
c        Mark Taylor & HG  (11/09/94 bootstrap update, and kill 4,1 mode)
c                            7/15/96 cleanup for Carlos
c   
c   For details of the theory see
c   
c   1.  M. Taylor, "A High Performance Spectral Code for 
c       Nonlinear MHD Stability,"  Ph.D. Thesis, New York University, 1992.
c   
c   2.  O. Betancourt, "BETAS, a spectral code for three-dimensional 
c       magnetohydrodynamic equilibrium and nonlinear stability calculations,"
c       Comm. Pure Appl. Math. 41, 551--568, 1988.  
c   
c   3.  F. Bauer, O. Betancourt, P. Garabedian,   
c      "Magnetohydrodynamic Equilibrium and Stability of Stellarators,"
c       Springer-Verlag, New York, 1984.
c   
c   
c                          MHD Equations
c   
c   In the following equations, s*pi is the toroidal flux.  s=const.
c   defines a nested family of tori with s=1 the outer wall and s=0
c   the magnetic axis.  On each torus, u and v represent poloidal and
c   toroidal angles. v is chosen to be the true toroidal angle in 
c   cylindrical coordinates x,v,y, while the choice of u depends on the 
c   parametrization of the outer wall.  On each Poincare section 
c   v=const. x and y represent Cartesian coordinates. 
c   
c   B    = grad(PSI') cross grad(s)
c   PSI' = pi*(u - iota*v) + PSI      where PSI is single valued
c   ENER = [.5*(B dot B) - p(s)] d(volume)
c   
c   Transform to flux variables (s,u,v) with unknowns PSI,R,x0 and y0:
c   
c   x = x0 + R*x2   where x2=x1-x0
c   y = y0 + R*y2   where y2=y1-y0
c   x1,y1    parametrization for plasma boundary
c   x0,y0    magnetic axis
c   
c   B  = ( PSI'u*Tv - PSI'v*Tu ) / JD
c   Tu = [ xu, yu, 0]
c   Tv = [ xv, yv, JFAC]
c   Ts = [ xs, ys, 0]
c   BS = B dot Ts 
c   BV = B dot Tv 
c   BU = B dot Tu 
c   
c   ENER  = [.5*(B dot B) - p(s)] * JFAC * D * hs * hu * hv
c   LPSI  = ( BVu - BUv ) 
c   LR    = [ (BUs-BSu)*PSIv - pp*JFAC*D - (BVs-BSv)*PSIu ] / Rs 
c   Liota = -curr(s) / s  
c   LX    = integral [ -pole*yu ] du  at the first flux surface s=hs
c   LY    = integral [  pole*xu ] du  at the first flux surface s=hs
c   
c   pole    = TuxTv(k,j) /(JFAC(k,j) * D(k,j,inext)**2)
c   curr(s) = the net toroidal current = integral BU du*dv
c   iota    = rotational transform = der. of poloidal flux w.r.t.
c             toroidal flux
c   H       = y2u*x2 - x2u*y2
c   JFAC    = zfac*(1+ep*x) = Jacobian (e1,e2,e3)/(x,v,y)
c   D       = Jacobian(x,v,y)/(s,v,u)=Jacobian(x,y)/(s,u)=xs*yu-xu*ys=
c           =.5*(R**2)s*H
c   JFAC*D  = Jacobian (e1,e2,e3)/(s,v,u)  (e1,e2,e3 are the R3
c             Cartesian coordinates)
c   ep      = inverse aspect ratio
c   Q       = number of field periods
c   zfac    = circumference of 1 field period = pi2/(ep*Q)   if ep<>0
c   zfac    = length of 1 field period = pi2*zlen/Q          if ep=0
c   zlen    = length of cylinder in units of the plasma radius * pi2,
c             if ep=0
c   rfac    = scales R equation
c   psifac  = scales PSI equation
c   
c   Artificial time second order Richardson iterative scheme
c   
c     a*RFtt +   e*RFt = M1**(-1) LRF
c   a*PSIFtt + e*PSIFt = M2**(-1) LPSIF
c   a*x0Ftt  +  e*x0Ft = LXF
c   a*y0Ftt  +  e*y0Ft = LYF
c   
c   e=safi*(a/ht)*w2    for R and PSI      
c   a,e                 given for the axis
c   
c   The F suffix denotes the Fourier transforms of the root variable.
c   The t subscript denotes derivatives with respect to artificial time
c   M1 = linear tridiagonal approximation of the principal part of LRF
c   M2 = linear diagonal approximation of the principal part of LPSIF
c   w2 = An estimate of the decay rate of the L2 error
c   
c   
c   
c                          GLOSSARY OF INPUT VARIABLES
c   
c   
c   ns               The number of intervals in the s (radial) direction
c                    sigma=i/ns where hsig=1/ns and i=0..ns and
c                    s=sigma**alpha.
c                    Suggested value 15.
c   
c   nuf              The maximum index to be used for the Fourier series
c                    in u.
c                    Fourier coefficients A(0)..A(2*nuf+1) are stored as
c                    follows
c   
c                    for m = 0..nuf:
c                    A(m)      = real part of exp(i*pi2*m*u) coefficient
c                    A(m+nuf+1) = imaginary part of exp(i*pi2*m*u)
c                    coefficient
c   
c                    Suggested value 8 for equilibrium runs, 2m+2 for
c                    stability runs with a mode number m.
c              
c                    
c   nvf              The maximum index to be used for the Fourier series
c                    in v.
c                    See nuf above.  Suggested value on the order of nuf.
c                    It is important not to have a lopsided mesh in u
c                    and v.
c   
c   
c   alpha            The mesh (sigma) is spaced to approximate s=sigma**
c                    alpha
c                    alpha = 1  mesh equally spaced in s (flux)
c                    alpha = 2  mesh equally spaced in radius
c                    alpha > 2  mesh clustered near axis
c                    alpha < 2  mesh clustered near wall
c                    Suggested value 1.  
c   
c   A                Aspect ratio.  Ratio of minor radius to major radius.  
c   
c   
c   Q                Number of field periods in the device.
c   
c   Qext             Number of field periods to extend the solution to
c                    and run numext more iterations.
c   
c   irot             Rotate coordinate system.  Make the change of
c                    variable from (uwall,v) to (u,v) where
c                    uwall=u+irot*v, for v=0..1 in 1 field period.
c                    Suggested value 0, unless the case has a large
c                    helical excursion where one should use 1.
c                    Irot changes the parametrization of the outer wall
c                    and allows us to exploit helical symmetry.  
c                    The only formulas affected are those which have an
c                    explicit u dependence given. 'hu' does not change.  
c                    The input for the outer wall, zoning and iota are 
c                    assumed to be given in the (uwall,v)
c                    coordinate system so they need not be changed when 
c                    irot is changed.  
c   
c                    Iota is changed by the code to account for the 
c                    change in the representation of PSI' from 
c                       PSI' = pi*(uwall - iota*v) + PSI(uwall,v)
c                    to 
c                       PSI' = pi*(u - (iota-irot)*v) + PSI(u,v).
c                    
c                    If we are running Qext field periods
c                    instead of 1, irot and iota are multiplied by Qext.
c   
c   
c   Zmn              Zoning parameters.  To redistribute the mesh so
c                    it can be made equally spaced in u. For each Zmn, 
c   
c                         uwall = u + sum Zmn * sin(pi2*(m*u-n*v))/pi2
c                                     m,n
c   
c                    where the outer wall is assumed to be given in the 
c                    (uwall,v) coordinate system. Subroutine wall
c                    encodes this into (u,v) coordinates.
c   
c   
c   m,n,delmn        The shape of the outer wall (x1,y1) in one
c                    field period is given by an arbitrary number of
c                    terms m,n and delmn.  The formula is
c                    
c                    x1+i*y1=exp(i*u)*sum[ delmn*exp(-i*(m*u-n*v)) ]
c                                           m,n
c   
c                    where the sum is over all terms present in
c                    the input file and a default additional term 
c                    giving a circular outer wall:
c                    m=0, n=0, del00=1.0.  
c   
c   
c   p0               Pressure profile  p(s)=p0*(1 - s**ypr)**xpr.
c   xpr,ypr
c   
c   
c   iota0,iota1      iota(s)=iota0 + iota1*s.  Initial value for iota in 
c                    one field period.
c   
c   stell            A logical variable which can be either T or F for
c                    true or false.  When true, iota is iterated to
c                    drive the toroidal current to zero.
c   
c   icont            Allows a run to be continued.
c                    Suggested value 0.  
c                    
c                    icont=1   The unknowns are initialized from the
c                              data in file 'restart', which comes from
c                              renaming the 'save' file produced by
c                              a previous run.  Qext must be 1 for
c                              all runs.
c                    icont=-1  Disable the production of the 'save' file. 
c                              Thus the run cannot be continued later.  
c                    
c   
c   numit            The number of iterations to perform.
c   
c   numext           Extend the solution to Qext field periods on the
c                    'numext' iteration.  Suggested value 0.
c                 
c   ic               Print every ic iteration.
c   
c   rays             T or F.  Plot rays if true.
c                    Suggested value F. 
c   
c   starth           The iteration number where the inhomogeneous 
c                    mode is turned on for stability test.  
c                    Set starth=0 to run only equilibrium.  
c   
c   stoph            The iteration number where the inhomogeneous 
c                    mode is turned off.  
c   
c   mh,nh            Determines which mode is used in the inhomogeneous
c                    term and which mode is tracked.
c   
c   mode             Size of the inhomogeneous mode.  The mode is
c                    LRh   = mode * s**2 *  (1-s)*cos(pi2*(mh*u - nh*v))
c                    LPSIh =-mode* s  *(1-4*s)*sin(pi2*(mh*u - nh*v))
c                    and if mh=1,
c                    LXh   =2*mode/5.0 * cos(pi2*nh*v)
c                    LYh   =2*mode/5.0 * sin(pi2*nh*v)
c                    See subroutine INHMODE.
c   
c   
c                              TABLE OF SUBROUTINES
c   
c   
c   nstab            program start up                                   379
c   
c   input            reads input data described in glossary             498
c   
c   wall             defines easily modified wall shape                 688
c   
c   init             initializes variables                              837
c   
c   control          main control of iteration                         1050
c   
c   elam             descent coefficients for acceleration scheme      1249
c   
c   iterate          fast iterative scheme                             1277
c   
c   residual         calculate residuals for output                    1546
c   
c   PDE              MHD equations and Mercier criterion               1696
c
c   deliota          rotational transform from bootstrap current       2414
c   
c   coord            coordinate system centered at the magnetic axis   2703
c   
c   inhmode          initializes inhomogeneous mode for stability test 2833
c   
c   extend           extend solution to several field periods for test 2966
c    
c   output           organize and print output                         3108
c   
c   restart          I/O for continuing a run                          3332
c   
c   plot             all plots using NCAR and real*4                   3435
c   
c   norm             L2 and maximum norms                              3728
c   
c   tri              tridiagonal solver for preconditioner             3753
c   
c   FTstore          trig functions for Fourier transform & entry's    3863
c   
c   FTV,FTU          entry to SGEMM for Fourier transform         4034,4054
c   
c   FTDV,FTDU        SGEMM fast multiplication for derivatives    4067,4082
c   
c   PDV,PDU          entry to SGEMM matrix multiplications & FT   4095,4105
c
c   second           CPU time for UNIX workstations                    4114
c   
c   
c   
c                         EXTERNAL SUBROUTINES
c   
c   NCAR             Graphics
c   SGEMM            Matrix multiply from level 3 BLAS; also DGEMM
c   second           CPU time
c   
c   This code was written to run on both CRAY supercomputers and unix
c   workstations.  The implicit real*8 declaration is used to give the
c   same precision on both types of machines.  To run this code on Sun
c   workstations where real*8 is double precession, the calls to external
c   routine SGEMM must be changed to DGEMM.  If possible, it is important
c   to use a version of SGEMM or DGEMM that has been optimized for the
c   particular machine the code is to be run on.  On the CONVEX C1, a
c   fast version of DGEMM is available in VECLIB.  The variables used
c   for plotting are declared real*4, which is what is required by NCAR
c   graphics on unix workstations.  On a CRAY, real*4 is ignored and
c   real*8 is used, as required by the CRAY implementation of NCAR
c   graphics.
c   
c   
c                          INTERNAL VARIABLES
c   
c   
c   In the code, the radial mesh is labeled by sigma with step size hsig.
c   The dependence of s on sigma is controlled by alpha (see glossary).
c   
c   Functions of s are either evaluated at nodes or centers.  The
c   location of the centers is chosen so that s at a center is the
c   average of s at the two adjacent mesh points.
c   
c   The s derivatives that appear are computed on the sigma mesh by
c   taking differences at adjacent mesh points (or centers) and dividing
c   by the difference in toroidal flux at those points.
c   
c   When computing integrals in the s direction on the sigma mesh, 
c   hs(i+.5)=(ds/dsigma)*dsigma = s(i+1)-s(i)
c   
c   Variables with the prefix 'av' represent an average over two centers
c   (or mesh points) to estimate the value of the original variable 
c   at a meshpoint (or center).
c   
c   Variables with the suffix 'F' represent the Fourier transform
c   of the original variable.  
c   
c   Variables with the indices k or j have values in u and v space,
c   where j=1..nu and k=1..nv
c   
c   Variables with the indices kf or jf have values in u and v Fourier
c   space, where jf=0..nuf2 and kf=0..nvf2
c   
c   Variables with the index i have values in s (or sigma) space where
c   i=0..ns.  i represents the ith surface for quantities evaluated at
c   mesh points, and surface between mesh point i and i+1 for quantities
c   at centers.  i=0 is the magnetic axis and i=ns is the outer wall.
c   
c   Some variables are dimensioned with the last index of 0:1, which is
c   used for several things:t to store information at two different time
c   levels (like the variables RF and PSIF), to store information at two
c   different values of s (such as Ru, which as the code loops through
c   all the mesh points its values are only needed on the previous two
c   mesh points) or for quantities like hs, s and srt, which are
c   functions of s and have values at both mesh points and centers.
c   
c
c
      program nstab
c     initialize program
c
c                   ARRAY STORAGE PARAMETERS
c 
c     ns1              must be at least as large as ns
c     nu1              must be at least as large as nu            
c     nv1              must be at least as large as nv
c     nuf1             must be at least as large as nuf
c     nvf1             must be at least as large as nvf
c     Qd               must be at least as large as Qext if numext > 0
c 
c     These values are often fixed at 
c 
c        ns1 x nuf1 x nvf1 = 22 x 16 x 12
c                     Q1 = 1
c                   
c     which uses less that 1Mw of memory on the CRAY/C.
c     For larger meshes, they must be changed and the code recompiled.
      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2
c
      open(unit=5,file='nstabh',form='formatted')   ! Input file
      open(unit=7,file='nstabout',form='formatted') ! Output file
      open(unit=21,file='nstabplt',form='formatted')! Plot file
      call input
      call init
      if (icont.eq.1) call restart(1)
c     call opngks

c     solve the MHD equations
      call control

      call output
      call plot
c     call plotit(0,0,0)
c     call clsgks

c     save results in a file to allow run to be continued
      if (icont.ne.-1) call restart(0)
      close(5)
      close(7)
      end



      subroutine input
c     input variables described in glossary
c
      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2
c

      character*132 inline
      do 1 m=-nufd,nufd
      do 1 n=-nvfd,nvfd
         zone(m,n)=0
1        del(m,n)=0

      read(5,*)
      read(5,2001) ns,nuf,nvf,Q,Qext,irot,alpha,A
      nu=3*nuf
      nv=3*nvf
      nvfAX=0.51*nvf         ! normally 0.76*nvf
      zlen=0.0
      if (A.ne.0) ep=1/A
      if (A.eq.0) ep=0
      numdel=0
      read(5,2020) inline
      if((index(inline,'p0').ne.0).or.(index(inline,'P0').ne.0))goto100
90    read(5,2020) inline
      if((index(inline,'p0').ne.0).or.(index(inline,'P0').ne.0))goto100
      read(inline,2008) m,n,del(m,n),zone(m,n)
      if (m.ge.nufd) stop 'mdel > nufd'
      if (n.ge.nvfd) stop 'mdel > nvfd'
      numdel=numdel+1
      goto 90
100   continue
      read(5,2004) p0,xpr,ypr,iota0,iota1,stell,rays,icont
      zpr=1.0
      eiota=50.0             !     normally  20.0

c     aR,aPSI control CFL time step of iterative scheme, emin controls descent
c     larger values slower, but better in hard cases and on fine grids

      aR=5.0        !  normally  3.0
      aPSI=5.0      !  normally  3.0
      emin=0.5      ! normally emin= 0.2      07/09/06
      nav=50
      safi=2.0
      aAX=0.0
      eAX=0.4      !  normally  0.4
      read(5,*)
      read(5,2009) numit,numext,starth,stoph,mh,nh,ic,mode
      nprint=12
      lthick=1.0
      bndj=300.0
      fixAX=0.0
      artvR=0.001
      artvPSI=0.001
      artvAX=5.0        !  normally  5.0
      nboot=0
      bootit=1000
      read(5,2020,err=110,end=110) inline


110   write(7,1000) 'ns','nuf','nvf','Q','Qext','irot','alpha','A'
      write(7,1001) ns,nuf,nvf,Q,Qext,irot,alpha,A
      if (numdel.ge.1) then
         write(7,*)
         write(7,1000) 'm','n','DELmn','Zmn'
      endif
      do 150 m=-nufd,nufd
      do 150 n=-nvfd,nvfd
150      if (del(m,n).ne.0 .or. zone(m,n).ne.0)
     $   write(7,1008) m,n,del(m,n),zone(m,n)
      write(7,*)
      write(7,1000) 'p0','xpr','ypr','iota0','iota1','stell',
     $       'rays','icont'
      write(7,1004) p0,xpr,ypr,iota0,iota1,stell,rays,icont
      write(7,*)
      write(7,1000) 'numit','numext','starth','stoph','mh','nh',
     $        'ic', 'mode'
      write(7,1006) numit,numext,starth,stoph,mh,nh,ic,mode

      ht=1.0
      pd1=.01
      pd2=.01
      pd3=.01



c     check storage or invalid input
      if (Qext.lt.1 .or. numext.le.0 .or. numit.lt.numext) Qext=1
      if (starth.gt.stoph) starth=0
      if (starth.le.0) stoph=0
      if (2*nvf+1.gt.nv) stop '2*nvf+1 > nv'
      if (2*nuf+1.gt.nu) stop '2*nuf+1 > nu'
      if (nuf .gt. nufd) stop 'nuf > nuf1'
      if (Qext*nvf .gt. nvfd) stop 'Qext*nvf > nvf1'
      if (nvfAX .gt. nvf) stop 'nvfAX > nvf'
      if (ns .lt. 2) stop '2 > ns'
      if (ns .gt. nsd) stop 'ns > ns1'
      if (nu .gt. nud) stop 'nu > nu1'
      if (Qext*nv .gt. nvd) stop 'Qext*nv > nv1'

1000  format(5x,10a8)
1001  format(5x,6i8,2f8.2)
3000  format(5x,5a8)
3001  format(5x,5i8)
1004  format(5x,f8.3,2f8.2,2f8.3,2l8,i8)
1006  format(5x,7i8,f8.4)
1008  format(5x,2i8,2f8.3)
2001  format(6i8,2f8.3)
2004  format(5f8.3,2l8,i8)
2008  format(2i8,2f8.3)
2009  format(7i8,f8.3)
2020  format(a132)
      end



      subroutine wall(u,v,xw,yw,xwu,ywu,xwv,ywv)
      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2
c
c
c Define outer wall and the point (u,v), where u and v range from
c 0 to 1.  
c
c The shape of the outer wall (xw,yw) in one field period is determined
c by an arbitrary number of input lines of the form m,n,delmn.
c 
c       xw + i*yw = exp(i*u) * sum[ delmn * exp(-i*(m*u - n*Qrun*v)) ]
c                              m,n
c
c where i=sqrt(-1) and Qrun is the number of field periods the calculation
c is going to be performed in.  
c
c The term 1.0*exp(i*u) corresponding to a circular outer wall is
c added by default.  
c
c This subroutine can be easily modified to allow any type of formula
c for the outer wall.  Only the quantities xw,yw, their u derivatives
c xwu,ywu and their v derivatives xwv,ywv need to be coded.  
c 


      vwall=v

c     rotation by irot
      uirot=u+irot*v
      uirotu=1
      uirotv=irot
      
c     rezone   
      uwall=uirot
      uwallu=uirotu
      uwallv=uirotv
      do 80 m=-nufd,nufd
      do 80 n=-nvfd,nvfd
         if (zone(m,n).ne.0) then
            arg=pi2*m*uirot - pi2*n*Qrun*vwall
            argu=pi2*m*uirotu
            argv=pi2*m*uirotv - pi2*n*Qrun
            uwall = uwall +      zone(m,n)*sin(arg)/pi2
            uwallu=uwallu + argu*zone(m,n)*cos(arg)/pi2
            uwallv=uwallv + argv*zone(m,n)*cos(arg)/pi2
         endif
80    continue

      xw = 0
      yw = 0
      xwu= 0
      ywu= 0
      xwv= 0
      ywv= 0
      do 90 m=-nufd,nufd
      do 90 n=-nvfd,nvfd
         if (del(m,n).ne.0) then
            m2=1-m
            arg=pi2*m2*uwall + pi2*n*Qrun*vwall
            argu=pi2*m2*uwallu 
            argv=pi2*m2*uwallv + pi2*n*Qrun
            xw = xw + del(m,n)*cos(arg)
            yw = yw + del(m,n)*sin(arg)
            xwu=xwu - del(m,n)*sin(arg)*argu
            ywu=ywu + del(m,n)*cos(arg)*argu
            xwv=xwv - del(m,n)*sin(arg)*argv
            ywv=ywv + del(m,n)*cos(arg)*argv
         endif
90    continue
10    continue
      end



      subroutine init
c     initialize values for RF, PSIF, x0F, y0F, iota, s, p
c
c
      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2
c
      real*8 sn,sc,p1,p2

c     initialize constants
      nfinal = numit + bootit*nboot
      pmod=2*nfinal/(numpd+0)  + 2     !  normally  +1)
      nump=0
      pi=4*atan(1d0)
      pi2=2*pi
      hsig=1d0/ns
      hv=1d0/nv
      hu=1d0/nu
      nuf2=2*nuf+1
      nvf2=2*nvf+1
      Qrun=1
      nh=nh-irot*mh      
      if (ep.ne.0) zfac=pi2/(ep*Q)
      if (ep.eq.0) zfac=pi2*zlen/Q
      tnew=1
      told=0
c     scale protection coefficients
      pd1=pd1/(2*zfac)**2
      pd2=pd2*(pi2/zfac)**2
      call FTstore

c     incorporate default circular outer wall into the delmn array
c     unless the user has input a del(0,0)
      if (del(0,0).eq.0) del(0,0) = 1

c     outer wall quantities from del(m,n) 
      do 100 j=1,nu
      do 100 k=1,nv
         u=(j-1)*hu
         v=(k-1)*hv
100      call wall
     $   (u,v,x1(k,j),y1(k,j),x1u(k,j),y1u(k,j),x1v(k,j),y1v(k,j))

c     initialize axis position at both time levels
      do 350 k=1,nv
         x0(k)=0
         y0(k)=0
         v=pi2*((k-1)*hv)
         do 340 n=-nvfAX,nvfAX
            x0(k)=x0(k)+del(1,n)*cos(Qrun*n*v)
            y0(k)=y0(k)+del(1,n)*sin(Qrun*n*v)
            x0(k)=x0(k)+del(-1,n)*cos(Qrun*n*v)
            y0(k)=y0(k)-del(-1,n)*sin(Qrun*n*v)
340      continue
350   continue
      do 360 tuse=0,1
      call FTV(1,x0,x0F(0,tuse))
      call FTV(1,y0,y0F(0,tuse))
      do 355 kf=0,nvf2
         n=mod(kf,nvf+1)
         if (n.gt.nvfAX) y0F(kf,tuse)=0
         if (n.gt.nvfAX) x0F(kf,tuse)=0
355   continue
360   continue

c     coordinate system
      call coord

c     initialize mesh
      do 400 i=0,ns
         s(i,0)=(i*hsig)**alpha
         srt(i,0)=sqrt(s(i,0))
         if (i.ne.0) s(i-1,1)=.5*(s(i-1,0)+s(i,0))
         if (i.ne.0) srt(i-1,1)=sqrt(s(i-1,1))
400   continue
      do 405 i=0,ns-1
         if (i.ne.0) hs(i,0)=s(i,1)-s(i-1,1)
         hs(i,1)=s(i+1,0)-s(i,0)
405   continue

c     initialize flux
      do 450 i=0,ns-1     ! iota formula
 450     iota(i)=iota0 + iota1*s(i,1) - irot
c450     iota(i)=iota0 + iota1*(0.60 - s(i,1))**2 - irot     !!
c450     iota(i)=iota0 + iota1*(0.61 - s(i,1))**2 - irot     !!
c450     iota(i)=iota0 + iota1*(0.50 - s(i,1))**2 - irot     !!

c     initialize R at both time levels
      do 200 tuse=0,1
      do 200 i=0,ns
         do 210 j=1,nu
         do 210 k=1,nv
            tmp1(k,j)=srt(i,0)*rwall(k,j)
210      continue
         call FTV(2,tmp1,ftwork)
         call FTU(2,ftwork,RF(0,0,i,tuse))
200   continue

c     initialize PSI at both time levels
      do 300 tuse=0,1
      do 300 i=0,ns-1
         do 310 j=1,nu
         do 310 k=1,nv
            tmp1(k,j)=0
310      continue
         call FTV(2,tmp1,ftwork)
         call FTU(2,ftwork,PSIF(0,0,i,tuse))
         PSIF(0,0,i,tuse)=0
         PSIF(nvf+1,0,i,tuse)=0
         PSIF(0,nuf+1,i,tuse)=0
         PSIF(nvf+1,nuf+1,i,tuse)=0
300   continue


c     initialize pressure
      do 500 i=0,ns-1
         sc=s(i,1)
         p(i)=p0*(1-zpr*sc**ypr)**xpr
500   continue
c     pp(i) is p' at mesh points 1..ns-1
      do 510 i=1,ns-1
         sn=s(i,0)
         p1=xpr*(1-zpr*sn**ypr)**(xpr-1) 
         p2= (-ypr*zpr*sn**(ypr-1))
         pp(i)=p0*p1*p2
510   continue
c     ppout is p' at centers for output only
      do 520 i=0,ns-1
         sc=s(i,1)
         p1=xpr*(1-zpr*sc**ypr)**(xpr-1) 
         p2= (-ypr*zpr*sc**(ypr-1))
         ppout(i)=p0*p1*p2
520   continue


      do 521 i=0,ns-1
         curr0(i) = 0.
 521  continue
      end



      subroutine control
c     main control of iteration
c
c
      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2
c
      parameter(navd=500)
      real*8 w(navd),e2(navd)
      character*1 char1
      logical iprint,iplot
      negjac=.false.
      if (nav.gt.navd) stop 'nav > navd'
c     initialize acceleration scheme
      errL2R=1
      errL2PSI=1
      errL2AX=1
      eav=1
      eR=aR*safi*eav
      ePSI=aPSI*safi*eav
      do 100 i=1,nav
         w(i)=0
         e2(i)=eav**2
100   continue

c     do nfinal iterations
c     call second(tstart)
      call cpu_time(tstart)
      do 200 it=1,nfinal
         char1=' '
         if (Qext.gt.1 .and. numext.eq.it) then     
c           extend to Qext field periods
            call extend
         endif

c        add inhomogeneous term for iterations between starth and stoph
         inhON=.false.
         if (it.eq.starth) call inhmode
         if ((it.ge.starth) .and. (it.le.stoph)) inhON=.true.
         if (inhON) char1='h'

c        compute MHD equations
         if ((it.ge.numit).and.mod(it-numit,bootit).eq.0) then
            call PDE(.true.)
         else
            call PDE(.false.)
         endif
         iprint=.false.
         iplot=.false.
         if ((it.eq.1)  .or. (it.eq.nfinal)) iprint=.true.
         if (mod(it,ic).eq.0) iprint=.true.
         if ((negjac) .or. (jac.gt.bndj)) iprint=.true.
         if (mod(it,pmod).eq.0 .or. it.eq.nfinal) iplot=.true.
c        compute residuals for output or plots
         if (iprint .or. iplot) call residual
         if (negjac .or. jac.gt.bndj) goto 150

c        take one time step in the iteration of R, PSI, and the axis location
         oldL2R   = errL2R
         oldL2PSI = errL2PSI
         call iterate

c        estimate error decay rate 
         xmean1=errL2R*errL2PSI
         xmean0=oldL2R*oldL2PSI
         w(nav)=(xmean1  - xmean0)  /(  ht*(xmean0+xmean1))
         if (it.gt.nav) then
            call elam(nav,eav,e2,w,ht)
            eR   = max(aR*safi*eav,emin)
            ePSI = max(aPSI*safi*eav,emin)
         endif

150      continue
c        iteration printout.  The values given below are printed for the 
c        first, last and every 'ic' iteration.
         if (it.eq.1) then
            write(7,*)
            write(7,*)
            write(7,1010) ' iter','  energy   ','  Jac  ratio',
     $    '     mode   ',
     $    '   max R err','    Axis err',
     $    '       eR   '

            write(7,*)
         endif
         if (iprint) write(7,1000) it,char1,ener,jac,
     $        modeval,
     $        errR,errAX,
     $        eR

1010     format(a8,1x,a12,99a12)       ! normally  5,
1000     format(i8,a1,f12.8,7e12.3)       ! normally  5,

c        store quantities for plotting        
         if (iplot) then
            nump=nump+1
            errp(nump,1)=errR2
            errp(nump,2)=errPSI2
            errp(nump,3)=errAX2
            errp(nump,4)=errI 
            errp(nump,5)=model2
         endif
         if (negjac) goto 300
         if (jac.gt.bndj) goto 400
200   continue
      go to 900

300   continue
      write(7,*)
      write(7,*) 'negative jacobian.  iteration terminated'
      goto 900

400   continue
      write(7,*)
      write(7,*) 'jacobian exceeds bound.  iteration terminated'
      goto 900

900   continue
c     call second(tstop)
      call cpu_time(tstop)
      write(7,*)
      write(7,9000) 'CPU time: ',(tstop-tstart)/60,' minutes.'
9000  format(a10,f8.2,a10)
      end



      subroutine elam(nav,eav,e2,w,ht)
c     Acceleration scheme
      implicit real*8 (a-h,o-z)
      real*8 w(nav),e2(nav)       
      integer t1
      enew=0
      do 120 t1=2,nav-1
         enew=enew + (w(t1+1)-w(t1))/ht
     $             + (e2(t1+1)-e2(t1-1))/(2*ht)
     $             + (.5*(w(t1+1)+w(t1)))**2
     $     + .5*(w(t1+1)+w(t1)) * .25*(e2(t1-1)+2*e2(t1)+e2(t1+1))
120   continue
      enew=abs(enew)/(nav-2)

      do 130 t1=1,nav-1
         w(t1)=w(t1+1)
         e2(t1)=e2(t1+1)
130   continue
      e2(nav)=enew

      eav=0
      do 140 t1=1,nav
140      eav=eav+e2(t1)
      eav=sqrt(eav/nav)
      end



      subroutine iterate
c     Applies the preconditioner and then takes one second order
c     Richardson time step for R, PSI and the axis equations
c     with the values of the acceleration coefficients 
c     determined previously in subroutine CONTROL.
c
c
      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2
c
      real*8 err,a(nsd),b(nsd),c(nsd),m2,n2,tmp,pc,L2sq
      real*8 a1,a2,a3,errx,erry

c     precondition axis 
      do 175 kf=0,nvf2
         n=mod(kf,nvf+1)
         LXF(kf)=LXF(kf)/(1000*max(1,n))
         LYF(kf)=LYF(kf)/(1000*max(1,n))
175   continue
      
c     precondition R and PSI equations 
      do 200 jf=0,nuf2
      m = mod(jf,nuf+1)
      m2= pi2**2 * m**2
      do 200 kf=0,nvf2
         n = mod(kf,nvf+1)
         n2= pi2**2 * n**2

c        PSI equation spectral preconditioner
         if (m.ne.0 .or. n.ne.0) then
         do 205 i=0,ns-1
            pc=m2*M2uu(i) + n2*M2vv(i) 
            if (pc.lt.0) write(7,*) 'PSI pc < 0'
205         LPSIF(kf,jf,i)=LPSIF(kf,jf,i)/pc
         endif

c        R combined tridiagonal and spectral preconditioner 
c        from coefficients M1uu,M1vv,M1ss.  
c        a(i),b(i) and c(i) are the lower diagonal, diagonal and 
c        upper diagonal entries of the preconditioner, which represents
c        the operator M1ss(s)*(sqrt(s)*R)ss / sqrt(s)
         do 210 i=1,ns-1
            pc=m2*M1uu(i) + n2*M1vv(i) 
            if (pc.lt.0) write(7,*) 'R pc < 0'
            tmp=.5*(M1ss(i-1)+M1ss(i))
            a(i)=-tmp*srt(i-1,0)/(hs(i,0)*hs(i-1,1)*srt(i,0))
            c(i)=-tmp*srt(i+1,0)/(hs(i,0)*hs(i,1)*srt(i,0))
            b(i)=2*tmp/hs(i,0)**2 + pc
210      continue

c        precondition LRF.  LRF is replaced by M**(-1) * LRF
c        where M denotes the tridiagonal preconditioner computed above
c        and stored in the arrays a(i),b(i) and c(i).  
         a(1)=0
         c(ns-1)=0
c     11/09/94  More of H1 fix for HG getting rid of LRF when R=0!
c     LRF(2,3,1)=0
c     LRF(nvf+3,nuf+4,1)=0
c        LRF(3,4,1)=0
c        LRF(nvf+4,nuf+5,1)=0
      LRF(3,nuf+4,1)=0
      LRF(nvf+4,3,1)=0
c        LRF(4,5,1)=0
c        LRF(nvf+5,nuf+6,1)=0
        call tri(a,b,c,LRF,kf,jf,nvf2d,nuf2d,nsd)  !(Don't comment this)
200   continue
     
c     iterate R 
c     RF(tnew),PSIF(tnew) are the values at time it-1
c     RF(told),PSIF(told) are the values at time it-2
c     overwrite the values of RF and PSIF at it-2 with values at time it.  
c     2nd order Richardson   a*Rtt + e*Rt = L(R)
      a1 = ht**2 / (aR+eR*ht)
      a3 = -aR/(aR+eR*ht)
c     first iteration must be steepest descent
      if (it.eq.1) a3=0
      a2 = 1 - a3
      do 300 i=1,ns-1
      do 300 jf=0,nuf2
      do 300 kf=0,nvf2
         RF(kf,jf,i,told) =
     $      a1*LRF(kf,jf,i) +
     $      a2*RF(kf,jf,i,tnew) +
     $      a3*RF(kf,jf,i,told)
300   continue
c     11/09/94  Put R(3,1)=0 at first surface to make H1 for HG converge!
c     RF(2,3,1,told)=0
c     RF(nvf+3,nuf+4,1,told)=0
c     LRF(2,3,1)=0
c     LRF(nvf+3,nuf+4,1)=0
c      RF(3,4,1,told)=0
c      RF(nvf+4,nuf+5,1,told)=0
c      LRF(3,4,1)=0
c      LRF(nvf+4,nuf+5,1)=0
      RF(3,nuf+4,1,told)=0
      RF(nvf+4,3,1,told)=0
      LRF(3,nuf+4,1)=0
      LRF(nvf+4,3,1)=0
c      RF(4,5,1,told)=0
c      RF(nvf+5,nuf+6,1,told)=0
c      LRF(4,5,1)=0
c      LRF(nvf+5,nuf+6,1)=0

c     iterate PSI
c     2nd order Richardson 
      a1 = ht**2 / (aPSI+ePSI*ht)
      a3 = -aPSI/(aPSI+ePSI*ht)
      if (it.eq.1) a3=0
      a2 = 1-a3 
      do 301 i=0,ns-1
      do 301 jf=0,nuf2
      do 301 kf=0,nvf2
         PSIF(kf,jf,i,told)=
     $      a1*LPSIF(kf,jf,i) +
     $      a2*PSIF(kf,jf,i,tnew) +
     $      a3*PSIF(kf,jf,i,told)
301   continue

c     iterate axis    
c     2nd order Richardson
      a1 = ht**2 / (aAX+eAX*ht)
      a3 = -aAX/(aAX+eAX*ht)
      if (it.eq.1) a3=0
      a2 = 1 - a3
      if (it.ge.1000*fixAX .and. fixAX.ne.0) then
c        freeze axis
         a1=0
         a2=1
         a3=0
      endif

      do 430 kf=0,nvf2
         n=mod(kf,nvf+1)
         if (n.le.nvfAX) then
            x0F(kf,told)=a1*LXF(kf) + a2*x0F(kf,tnew) + a3*x0F(kf,told)
            y0F(kf,told)=a1*LYF(kf) + a2*y0F(kf,tnew) + a3*y0F(kf,told)
         else
            x0F(kf,told)=0
            y0F(kf,told)=0
         endif
430   continue

c     L2 error of preconditioned equations used by acceleration scheme
      errL2R=0
      errL2PSI=0
      do 250 i=1,ns-1
         call norm(err,L2sq,nloc,mloc,LRF(0,0,i),nvf2d,nvf2,nuf2)
250      errL2R=errL2R+L2sq*hs(i,0)*2

      do 251 i=0,ns-1
         call norm(err,L2sq,nloc,mloc,LPSIF(0,0,i),nvf2d,nvf2,nuf2)
251      errL2PSI=errL2PSI+L2sq*hs(i,1)*2

      call norm(errx,errL2AX,nloc,mloc,LXF,nvf2d,nvf2,0)
      call norm(erry,L2sq,nloc,mloc,LYF,nvf2d,nvf2,0)
      errL2AX=sqrt((errL2AX+L2sq)*2)
      errL2R=sqrt(errL2R)
      errL2PSI=sqrt(errL2PSI)

c     swap values at time it and it-1
      itmp=tnew
      tnew=told
      told=itmp

c     iterate iota
      errI=0
      do 400 i=0,ns-1
         errI=max(errI,abs(curr(i)/(Qrun*M3(i))))
c         if (stell) iota(i)=iota(i)-curr(i)/(eiota*M3(i))
         if (stell) iota(i)=iota(i)-(curr(i)-curr0(i))
     .                      /(eiota*M3(i))
400   continue

c     norm of mode we are tracking
      sgn=sign(1,nh)
      nhh=abs(nh)
      modeval=.5*(RF(nhh,mh,ns/2,tnew)+                            !
     $     sgn*RF(nhh+nvf+1,mh+nuf+1,ns/2,tnew))/srt(ns/2,0)       !
c     modeval=.5*(RF(nhh,mh+nuf+1,ns/2,tnew)-                      !
c    $     sgn*RF(nhh+nvf+1,mh,ns/2,tnew))/srt(ns/2,0)             !
      if (nh.eq.0 .or. mh.eq.0) modeval=2*modeval
      model2=0
      do 410 i=1,ns-1
         model2=model2 + (hs(i,0)/srt(i,0))*
     $        (.5*(RF(nhh,mh,i,tnew)+                              !
     $        sgn*RF(nhh+nvf+1,mh+nuf+1,i,tnew)))**2               !
c    $        (.5*(RF(nhh,mh+nuf+1,i,tnew)-                        !
c    $        sgn*RF(nhh+nvf+1,mh,i,tnew)))**2                     !
410   continue
      model2=sqrt(model2)
      if (nh.eq.0 .or. mh.eq.0) model2=2*model2
      end



      subroutine residual
c     max and L2 errors of unpreconditioned equations for output and plots
c     scale error by m=n=1 value of the preconditioner
c     multiply L2**2 errors by 2 since half of Fourier coefficients are 0
c
c
      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2
c
      real*8 L2sq
c     logical iprint
      errR=0
      errR2=0
      errPSI=0
      errPSI2=0
      errAX=0
      errAX2=0
c     11/09/94 Fix L2 error in R because of H1 fix to get HG convergence!
c     LRF(2,3,1)=0
c     LRF(nvf+3,nuf+4,1)=0
c      LRF(3,4,1)=0
c      LRF(nvf+4,nuf+5,1)=0
      LRF(3,nuf+4,1)=0
      LRF(nvf+4,3,1)=0
c      LRF(4,5,1)=0
c      LRF(nvf+5,nuf+6,1)=0
      do 150 i=1,ns-1
         call norm(err,L2sq,nloc,mloc,LRF(0,0,i),nvf2d,nvf2,nuf2)
         pc=.5*(M1ss(i-1)+M1ss(i)) + pi2**2*(M1uu(i) + Qrun**2*M1vv(i))
         err=err/pc
         L2sq=2*L2sq/pc
         if (err.gt.errR) then
            errR=err
            locR=i
            locRm=mod(mloc,nuf+1)
            locRn=mod(nloc,nvf+1) 
         endif
150      errR2=errR2+L2sq*hs(i,0)
      errR2=sqrt(errR2)
c     switch sign of locRn if error is in cos(m*u+n*v) term rather
c     than cos(m*u-n*v) term
      xx=.5*abs(LRF(locRn,locRm,locR)+LRF(locRn+nvf+1,locRm+nuf+1,locR))
      yy=.5*abs(LRF(locRn,locRm,locR)-LRF(locRn+nvf+1,locRm+nuf+1,locR))
      if (yy.gt.xx) locRn=-locRn
      locRn=locRn + irot*locRm

      do 151 i=0,ns-1
         call norm(err,L2sq,nloc,mloc,LPSIF(0,0,i),nvf2d,nvf2,nuf2)
         pc=pi2**2*(M2uu(i) + Qrun**2*M2vv(i))
         err=err/pc
         L2sq=2*L2sq/pc
         if (err.gt.errPSI) then
            errPSI=err
            locPSI=i
            locPSIm=mod(mloc,nuf+1)
            locPSIn=mod(nloc,nvf+1) 
         endif
151      errPSI2=errPSI2+L2sq*hs(i,1)*2
      errPSI2=0.5*sqrt(errPSI2)     !  normally  1.0
c     switch sign of locPSIn if error is in sin(m*u+n*v) term rather
c     than sin(m*u-n*v) term
      sinm=.5*abs(LPSIF(locPSIn,locPSIm+nuf+1,locPSI)-
     $         LPSIF(locPSIn+nvf+1,locPSIm,locPSI))
      sinp=.5*abs(LPSIF(locPSIn,locPSIm+nuf+1,locPSI)+
     $         LPSIF(locPSIn+nvf+1,locPSIm,locPSI))
      if (sinp.gt.sinm) locPSIn=-locPSIn
      locPSIn=locPSIn + irot*locPSIm
      call norm(errx,errL2AX,nloc,mloc,LXF,nvf2d,nvf2,0)
      call norm(erry,L2sq,nloc,mloc,LYF,nvf2d,nvf2,0)
      locAXn=mod(nloc,nvf+1)
      errAX=0.2*max(erry,errx)      !  normally 1.0
      errAX2=0.2*sqrt(errL2AX+L2sq)      !  normally 1.0

      end


      subroutine PDE(cbmn)
c     derivatives in MHD equations at mesh points
c     compute the axis residue at the magnetic axis i=0 
c     compute the R PDE at mesh points  i=1..ns-1
c     compute the PSI PDE at centers i=0..ns-1 
c
c     if cbmn=.true., compute Bmn's and Mercier in addition to the above.  
c     We do not want to do this time consuming process at every iteration.
c     However this cannot go into a seperate subroutine because all the
c     intermediate terms required are calcuated in subroutine PDE and
c     are not saved at every flux surface.  
c
c
c     For quantities evaluated at mesh points,  
c     i=0 is the magnetic axis, i=ns is the outer wall.
c     For quantities evaluated at centers, i is the surface between
c     mesh point surface i and mesh point surface i+1
c     
c     LPSI is computed on centers and uses data at mesh points i and i+1
c
c     LR is computed on mesh point i and uses data at centers i-1 and i
c
c     Thus all the intermediate quantities needed to compute LPSI and LR 
c     are only stored for two surfaces, indexed by 'ilast' and 'inext'.  
c
c     For example: 
c  
c     if BMESH(j,k,iuse) is evaluated at mesh points and
c     BCENT(j,k,iuse) is evaluated at centers, then 
c
c     BMESH(j,k,inext) = the values on mesh point surface i+1
c     BMESH(j,k,ilast) = the values on mesh point surface i  
c     
c     BCENT(j,k,inext) = values on center i (between mesh points i and i+1)
c     BCENT(j,k,ilast) = the values on center i-1
c
c     The variables x0,x2,x2u,x2v,y0,y2,y2u,y2v are used and are computed
c     by a call to coord
c 
c     also compute 
c        avb   = integral of .5*B*B  d(volume)
c        avp   = integral of p(i) d(volume)
c        vol   = integral of 1 d(volume)
c        ener  = avb - avp
c     These integrals are computed over all space and so are not defined
c     until PDE has been called with i=0..ns-1.
c


      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2
c
      real*8 mn,mx,jaci,meanRs,
     $ BUFv(0:nvf2d,0:nuf2d),BVFu(0:nvf2d,0:nuf2d),
     $ avR(nvd,nud),avRu(nvd,nud),avRv(nvd,nud),
     $ LR(nvd,nud),BVs(nvd,nud),avBS(nvd,nud),
     $ avBSv(nvd,nud),avPSIu(nvd,nud),avRs(nvd,nud),
     $ avJD(nvd,nud),avPSIv(nvd,nud),BUs(nvd,nud),
     $ avBSu(nvd,nud),JFAC2(nvd,nud),JFAC3,
     $ fac2,fac1,PHI(nvd,nud),PSI(nvd,nud),PHIF(0:nvf2d,0:nuf2d),
     $ BVF(0:nvf2d,0:nuf2d),BUF(0:nvf2d,0:nuf2d),
     $ grads2(nvd,nud,0:nsd)

      real*8 
     $     BMODkj(nvd,nud,0:nsd),
     $     BMODout(0:nufd,-nufd-nvfd:nufd+nvfd,0:nsd),
     $     JDns(nvd,nud,0:nsd),
     $     BUns(nvd,nud,0:nsd),BVns(nvd,nud,0:nsd),
     $     grads2b(0:nufd,-nvfd:nvfd)
      real*8 av1(0:nsd),av2(0:nsd),av3(0:nsd),av4(0:nsd)

      logical cbmn

      ilast=0
      inext=1
      jac=0
      avb=0
      avp=0
      vol=0
      ener=0
      do 170 j=1,nu
      do 170 k=1,nv
         Rv(k,j,ilast)=0
         Ru(k,j,ilast)=0
170      R(k,j,ilast)=0
      call coord

      do 1000 i=0,ns-1

c     compute R,Ru,Rv on flux surface i+1
      if (i+1.eq.ns) then
         do 180 j=1,nu
         do 180 k=1,nv
            R(k,j,inext) = rwall(k,j)
            Ru(k,j,inext)= rwallu(k,j)
            Rv(k,j,inext)= rwallv(k,j)
180      continue
      else
         call FTDU(-2,ftwork,RF(0,0,i+1,tnew))
         call FTV(-2,Ru(1,1,inext),ftwork)
         call FTU(-2,ftwork,RF(0,0,i+1,tnew))
         call FTV(-2,R(1,1,inext),ftwork)
         call FTDV(-2,Rv(1,1,inext),ftwork)
      endif

c     compute quantities on center i using values on mesh point i+1
c     calculated above, and values on mesh point i calculated previously

      s1=srt(i,0)
      s2=srt(i+1,0)
      sm=srt(i,1)
      do 210 j=1,nu
      do 210 k=1,nv
         Rs(k,j,inext) = (R(k,j,inext)-R(k,j,ilast))
     $                  / (2*srt(i,1)*(srt(i+1,0)-srt(i,0)))
         D(k,j,inext)=
     $        .5*H(k,j)*((R(k,j,inext)**2-R(k,j,ilast)**2)/hs(i,1))

         avR(k,j) = .5*(s2*R(k,j,inext)+s1*R(k,j,ilast))/sm
         avRu(k,j)= .5*(s2*Ru(k,j,inext)+s1*Ru(k,j,ilast))/sm
         avRv(k,j)= .5*(s2*Rv(k,j,inext)+s1*Rv(k,j,ilast))/sm
210   continue

      do 215 j=1,nu
      do 215 k=1,nv
         avxu(k,j)= avRu(k,j)*x2(k,j) + avR(k,j)*x2u(k,j)
         avyu(k,j)= avRu(k,j)*y2(k,j) + avR(k,j)*y2u(k,j)
         avxv(k,j)= x0v(k) + avRv(k,j)*x2(k,j) + avR(k,j)*x2v(k,j)
         avyv(k,j)= y0v(k) + avRv(k,j)*y2(k,j) + avR(k,j)*y2v(k,j)
         xs(k,j)  = Rs(k,j,inext)*x2(k,j)
         ys(k,j)  = Rs(k,j,inext)*y2(k,j)
         JFAC(k,j)= zfac*( 1 + ep * (x0(k) + avR(k,j)*x2(k,j)) )
215   continue

      do 220 j=1,nu
      do 220 k=1,nv
         TuTu(k,j)=avxu(k,j)**2 + avyu(k,j)**2
         TuTv(k,j)=avxu(k,j)*avxv(k,j) + avyu(k,j)*avyv(k,j)
         TvTv(k,j)=avxv(k,j)**2 + avyv(k,j)**2 + JFAC(k,j)**2
         TvTs(k,j)=avxv(k,j)*xs(k,j) + avyv(k,j)*ys(k,j)
         TuTs(k,j)=avxu(k,j)*xs(k,j) + avyu(k,j)*ys(k,j)
220   continue

      do 250 j=1,nu
      do 250 k=1,nv
         JD(k,j)=JFAC(k,j)*D(k,j,inext)
 250  continue

c     check for negative jacobian and compute jacobian ratio
      mx=0
      mn=1d10
      do 245 j=1,nu
      do 245 k=1,nv
         if (JD(k,j).gt.mx) mx=JD(k,j)
245      if (JD(k,j).lt.mn) mn=JD(k,j)
      if (mn.lt.0) then
         negjac=.true.
         return
      endif
      jaci=mx/(mn+1.d-6)
      jac=max(jac,jaci)


c     compute PSIu,PSIv on center i
      call FTDU(-2,ftwork,PSIF(0,0,i,tnew))
      call FTV(-2,PSIu(1,1,inext),ftwork)
      call FTU(-2,ftwork,PSIF(0,0,i,tnew))
      call FTDV(-2,PSIv(1,1,inext),ftwork)
      do 230 j=1,nu
      do 230 k=1,nv
         PSIu(k,j,inext)=PSIu(k,j,inext) + pi
         PSIv(k,j,inext)=PSIv(k,j,inext) - iota(i)*pi
230   continue

      avbi=0
      volp(i)=0
      do 260 j=1,nu
      do 260 k=1,nv
         BS(k,j,inext) = (PSIu(k,j,inext)*TvTs(k,j)
     $   - PSIv(k,j,inext)*TuTs(k,j) ) / JD(k,j)
         BV(k,j,inext) = (PSIu(k,j,inext)*TvTv(k,j)
     $   - PSIv(k,j,inext)*TuTv(k,j) )  / JD(k,j)
         BU(k,j,inext) = (PSIu(k,j,inext)*TuTv(k,j)
     $   - PSIv(k,j,inext)*TuTu(k,j) ) / JD(k,j)
         bsq(k,j,inext) = ( (PSIu(k,j,inext)**2)*TvTv(k,j)
     $   + (PSIv(k,j,inext)**2)*TuTu(k,j)
     $   - 2*PSIu(k,j,inext)*PSIv(k,j,inext)*TuTv(k,j)  )
     $   / JD(k,j)**2
         avbi=avbi + .5*bsq(k,j,inext)*JD(k,j)
         volp(i)=volp(i) + JD(k,j)
260   continue

      volp(i)=volp(i)*hu*hv
      avb=avb+avbi*hu*hv*hs(i,1)
      vol=vol+volp(i)*hs(i,1)
      avp=avp+p(i)*volp(i)*hs(i,1)
      ener=avb-avp

c     net toroidal current
      curr(i)=0
      do 300 j=1,nu
      do 300 k=1,nv
         curr(i)=curr(i)+BU(k,j,inext)
300   continue
      curr(i)=curr(i)*hu*hv


c     axis residue condition 
      if (i.eq.1) then
c     integrate reciprocal of Jacobian D defining residue at axis
c
c     In the special case when i=1:  
c
c     index    terms on mesh points      terms on centers
c
c     ilast    value on surface i=1      value of center  .5
c     inext    value on surface i=2      value on center 1.5
c                          
      scale1= 1 /(8*pi*zfac*srt(1,0)*s(1,0))
      do 401 k=1,nv
         LX(k)=0
         LY(k)=0
         do 400 j=1,nu
c           use filtered values of x0v, y0v
            xv= x0filv(k)
            yv= y0filv(k)
            xu= Ru(k,j,ilast)*x2(k,j) + R(k,j,ilast)*x2u(k,j)
            yu= Ru(k,j,ilast)*y2(k,j) + R(k,j,ilast)*y2u(k,j)
            JFAC3=zfac*(1+ep*x0(k))
            TuxTv=(xv*yu - yv*xu)**2  + (xu**2 + yu**2) * (JFAC3**2)
            D3=.5*H(k,j)*( R(k,j,ilast)**2/s(i,0))
            pole=scale1* TuxTv /(JFAC3 * D3**2)
            LX(k)=LX(k) - pole*yu*hu
            LY(k)=LY(k) + pole*xu*hu
400      continue
401   continue

      call FTV(1,LX,LXF)
      call FTV(1,LY,LYF)
      if (inhAX .and. inhON) then
         do 410 kf=0,nvf2
            LXF(kf)=LXF(kf) + lamAX*LXFh(kf)
410         LYF(kf)=LYF(kf) + lamAX*LYFh(kf)
      endif
      do 411 kf=0,nvf2
         n=mod(kf,nvf+1)
         if (n.gt.nvfAX) LXF(kf)=0
         if (n.gt.nvfAX) LYF(kf)=0
411   continue
      endif

c     scaling of R and PSI equations
      if (i.ne.0) rfac0=pi2*zfac/srt(i,0)
      rfac1=pi2*zfac/srt(i,1)
      psifac=zfac/pi

c     coefficients of PSI and iota preconditioner, part of R preconditioner
      M1ss(i)=0
      M2uu(i)=0
      M2vv(i)=0
      do 500 j=1,nu
      do 500 k=1,nv
         M1ss(i) = M1ss(i)+bsq(k,j,inext)*JFAC(k,j)*H(k,j)*
     $            .5*(R(k,j,inext)+R(k,j,ilast))
         M2uu(i) = M2uu(i) + TvTv(k,j)/JD(k,j)
         M2vv(i) = M2vv(i) + TuTu(k,j)/JD(k,j)
500   continue
      M1ss(i)=M1ss(i)*hu*hv/rfac1 + pd3
      M2uu(i)=M2uu(i)*hu*hv/psifac
      M3(i)=M2vv(i)*hu*hv
      M2vv(i)=M2vv(i)*hu*hv/psifac + pd2



c     PSI PDE 
      call FTV(2,BV(1,1,inext),ftwork)
      call FTDU(2,ftwork,BVFu)
      call FTDV(2,BU(1,1,inext),ftwork)
      call FTU(2,ftwork,BUFv)
      do 510 jf=0,nuf2
      do 510 kf=0,nvf2
         LPSIF(kf,jf,i)=(BVFu(kf,jf)-BUFv(kf,jf))/psifac
         if (inhPDE .and. inhON)
     $   LPSIF(kf,jf,i)=LPSIF(kf,jf,i) + lamPSI(i)*LPSIFh(kf,jf)
510   continue
c     remove the arbitrary function of s from PSI
      LPSIF(0,0,i)=0
      LPSIF(nvf+1,0,i)=0
      LPSIF(0,nuf+1,i)=0
      LPSIF(nvf+1,nuf+1,i)=0


      if (i.ne.0) then
c     R PDE and R preconditioner coefficients
      M1vv(i)=0
      M1uu(i)=0
      meanRs=0

      do 600 j=1,nu
      do 600 k=1,nv
         BVs(k,j)   = (BV(k,j,inext) - BV(k,j,ilast)) / hs(i,0)
         BUs(k,j)   = (BU(k,j,inext) - BU(k,j,ilast)) / hs(i,0)
         avRs(k,j)  = .5*(Rs(k,j,ilast)+Rs(k,j,inext))
         avBS(k,j)  = .5 * (BS(k,j,inext) + BS(k,j,ilast))
         avPSIu(k,j)= .5 * (PSIu(k,j,ilast) + PSIu(k,j,inext))
         avPSIv(k,j)= .5 * (PSIv(k,j,ilast) + PSIv(k,j,inext))
         JFAC2(k,j) = zfac*( 1+ep*(x0(k)+R(k,j,ilast)*x2(k,j)) )
         avJD(k,j)  = JFAC2(k,j)*.5*(D(k,j,ilast) + D(k,j,inext))
         tmp1(k,j)  = JFAC2(k,j)*H(k,j)*R(k,j,ilast)
         M1vv(i) = M1vv(i)+rscale(k,j)*avPSIu(k,j)**2 / tmp1(k,j)
         M1uu(i) = M1uu(i)+rscale(k,j)*avPSIv(k,j)**2 / tmp1(k,j)
         meanRs=meanRs + avRs(k,j)
600   continue
      meanRs = meanRs * hu*hv
      M1vv(i)= M1vv(i)*hu*hv/rfac0
      M1uu(i)= M1uu(i)*hu*hv/rfac0 + pd1
      call PDV(2,avBS,avBSv)
      call PDU(2,avBS,avBSu)

      
      do 610 j=1,nu
      do 610 k=1,nv
         LR(k,j)=(  (BUs(k,j)-avBSu(k,j))*avPSIv(k,j)
     $             -(BVs(k,j)-avBSv(k,j))*avPSIu(k,j)
     $             -pp(i)*avJD(k,j) ) * meanRs/(avRs(k,j)*rfac0)
610   continue
      call FTV(2,LR,ftwork)
      call FTU(2,ftwork,LRF(0,0,i))

c     inhomogeneous mode for stability test
      if (inhPDE .and. inhON) then
         do 650 jf=0,nuf2
         do 650 kf=0,nvf2
650         LRF(kf,jf,i)=LRF(kf,jf,i) + lamR(i)*LRFh(kf,jf)
      endif
      endif
c        /* Store stuff for Mercier calculation and
c           calculate sqrt(g)*sqrt(grads**2) on equilibrium grid */

         if (cbmn) then
         do 7100 j=1,nu
         do 7100 k=1,nv
            JDns(k,j,i)=JD(k,j)
            BVns(k,j,i) = BV(k,j,inext)
            BUns(k,j,i) = BU(k,j,inext)
            grads2(k,j,i)=
     $      (   (avxv(k,j)*avyu(k,j) - avyv(k,j)*avxu(k,j))**2
     $          + (avxu(k,j)**2 + avyu(k,j)**2) * (JFAC(k,j)**2) )
            grads2(k,j,i)= sqrt(grads2(k,j,i))
 7100    continue
         endif


c     move data on surface 'inext' to surface 'ilast' 
      itmp=ilast
      ilast=inext
      inext=itmp
c     loop back and do the next flux surface
1000  continue

c     add artificial viscosity to axis
      if (artvAX.gt.0.0) then
      do 1100 kf=0,nvf2
         n=mod(kf,nvf+1)
         if (n.le.nvfAX) then
            xn=real(n)/Qrun
            xnmax=real(max(nvfAX,1))/Qrun
            artv=artvAX*(xn**4)/(xnmax**3)
            LXF(kf)=LXF(kf) - artv*x0F(kf,tnew)
            LYF(kf)=LYF(kf) - artv*y0F(kf,tnew)
         endif
1100  continue
      endif

c     add artificial viscosity to R and PSI equations
      do 1110 jf=0,nuf2
      m = mod(jf,nuf+1)
      do 1110 kf=0,nvf2
         n = mod(kf,nvf+1)
         xm=real(m)/Qrun
         xn=real(n)/Qrun
         xmmax=real(max(nuf,1))/Qrun
         xnmax=real(max(nvf,1))/Qrun
         artv=(pi2**2)*((xm**2+xn**2)**2)/(xmmax**2+xnmax**2)
         if ((m.ne.0 .or. n.ne.0) .and. (artvPSI.gt.0.0)) then
            do 1105 i=0,ns-1
1105           LPSIF(kf,jf,i)=LPSIF(kf,jf,i) -
     $         artvPSI*artv*PSIF(kf,jf,i,tnew)/(100*(1+i**5))
         endif
         if (artvR.gt.0.0) then
            do 1106 i=1,ns-1
1106           LRF(kf,jf,i)=LRF(kf,jf,i) - 
     $         artvR*artv*RF(kf,jf,i,tnew)/(100*(1+i**5))
         endif
1110  continue

      if (cbmn) then
c     First Mercier loop: Quantities are evaluated at centers.

      do 1500 i=0,ns-1

c        Compute Bmn in flux coordinates

c        net poloidal current
         pcurr(i)=0
         do 700 j=1,nu
         do 700 k=1,nv
c           tmp1(k,j)=BVns(k,j,i)
c           tmp2(k,j)=BUns(k,j,i)
            pcurr(i)=pcurr(i)+BVns(k,j,i)
700      continue
         pcurr(i)=pcurr(i)*hu*hv

c        normalization factors
         fac2=curr(i)/pcurr(i)
         fac1=pi2/(1+iota(i)*fac2)

c        Fourier transform of BU,BV
         call FTV(2,BVns(1,1,i),ftwork)
c        call FTV(2,tmp1,ftwork)
         call FTU(2,ftwork,BVF)
         call FTV(2,BUns(1,1,i),ftwork)
c        call FTV(2,tmp2,ftwork)
         call FTU(2,ftwork,BUF)

c        PHIF from BV and BU
c          *vocl loop,nopreex
         do 710 jf=0,nuf
         do 710 kf=0,nvf
            if (jf.eq.0) then
               if (kf.eq.0) then
                  PHIF(kf,jf)=0.
                  PHIF(kf+nvf+1,jf+nuf+1)=0.
               else
                  PHIF(kf,jf)= BVF(kf+nvf+1,jf)/pi2/float(kf)
                  PHIF(kf+nvf+1,jf)= BVF(kf,jf)/pi2/float(kf)
           PHIF(kf,jf+nuf+1)= BVF(kf+nvf+1,jf+nuf+1)/pi2/float(kf)
           PHIF(kf+nvf+1,jf+nuf+1)= BVF(kf,jf+nuf+1)/pi2/float(kf)
               endif
            else
                  PHIF(kf,jf)=-BUF(kf,jf+nuf+1)/pi2/float(jf)
                  PHIF(kf,jf+nuf+1)= BUF(kf,jf)/pi2/float(jf)
          PHIF(kf+nvf+1,jf)=-BUF(kf+nvf+1,jf+nuf+1)/pi2/float(jf)
          PHIF(kf+nvf+1,jf+nuf+1)= BUF(kf+nvf+1,jf)/pi2/float(jf)
            endif
710      continue 

c        PSI,PHI from PSIF,PHIF
         call FTU(-2,ftwork,PHIF)
         call FTV(-2,PHI,ftwork)
         call FTU(-2,ftwork,PSIF(0,0,i,tnew))
         call FTV(-2,PSI,ftwork)
     

c        normalize PHI and PSI

         do 720 j=1,nu
         u=(j-1)*hu
         do 720 k=1,nv
            v=(k-1)*hv
            PSI(k,j)= (PSI(k,j) + pi*u - pi*iota(i)*v)/pi
            PHI(k,j)= (PHI(k,j)+curr(i)*u+pcurr(i)*v)/pcurr(i)
 720     continue

c          Bmn contains the coefficients of the Jacobian proportional
c          to 1/B**2 in flux coordinates. Note the change of sign
c          from the array JD which holds the negative of the Jacobian
c          of the equilibrium grid.

         facdel = 2.*hu*hv
         do 740 m=0,nuf
         do 740 n=-nvf,nvf
            grads2b(m,n) = 0.
            Bmn(m,n,i)=0
            do 730 j=1,nu
            do 730 k=1,nv               
               PHI2=fac1*(PHI(k,j) - fac2*PSI(k,j))
               PSI2=fac1*(PSI(k,j) + iota(i)*PHI(k,j))
               cs = cos(m*PSI2 - n*PHI2)
               Bmn(m,n,i)=Bmn(m,n,i) - JDns(k,j,i)*cs*facdel
               grads2b(m,n) =
     .                grads2b(m,n)+grads2(k,j,i)*cs*facdel
730         continue
740      continue
         do 7410 n=-nvf,nvf
            Bmn(0,n,i)   = 0.5*Bmn(0,n,i)
            grads2b(0,n) = 0.5*grads2b(0,n)
7410     continue
         do 7415 n=1,nvf
            Bmn(0,n,i) = Bmn(0,n,i) + Bmn(0,-n,i)
            Bmn(0,-n,i) = 0.
 7415    continue


c        Find Mod B coefficients for bootstrap calculation
c        Apply smoothing to ignore aliasing of high frequencies

         nufm = min(nuf*.5,12.0)      !  normally  nuf*.5
         nvfm = min(nvf*.5,12.0)      !  normally  nvf*.5
c         nufm = nuf*.90
c         nvfm = nvf*.90
c         nufm = nuf
c         nvfm = nvf
         do 7420 j=1,nu
            do 7420 k=1,nv
               BMODkj(k,j,i) = 0.
               do 7419 m=0,nufm
                  do 7419 n=-nvfm,nvfm
                     arg=pi2*(m*(j-1)*hu - n*(k-1)*hv)
                 cs = (cos(arg))*(1.-m**2/nufm**2)*(1.-n**2/nvfm**2)
                     BMODkj(k,j,i) = BMODkj(k,j,i) +
     &                  Bmn(m,n,i)*cs
 7419          continue
               BMODkj(k,j,i) = -BMODkj(k,j,i)/(pi*(pcurr(i)
     &              + iota(i)*curr(i)))
           BMODkj(k,j,i) = 1./sqrt(sqrt(BMODkj(k,j,i)**2 + 1.d-4))    !
 7420    continue
         
         do 7425 m=0,nuf
            do 7425 n=-nvf,nvf
               BMODmn(m,n,i) = 0.
               do 7425 j=1,nu
                  do 7425 k=1,nv
                     arg=pi2*(m*(j-1)*hu - n*(k-1)*hv)
                     cs = cos(arg)
                     BMODmn(m,n,i) = BMODmn(m,n,i) +
     &                 BMODkj(k,j,i)*cs*facdel
 7425    continue
         do 7426 n=-nvf,nvf
            BMODmn(0,n,i) = 0.5*BMODmn(0,n,i)
 7426    continue
         do 7427 n=1,nvf
            BMODmn(0,n,i) = BMODmn(0,-n,i) + BMODmn(0,n,i)
            BMODmn(0,-n,i) = 0.
 7427    continue

c           J dot B 
         Xmn(m,n,i) = 0.
         do 750 m=1,nuf
         do 750 n=-nvf,nvf
            Xmn(m,n,i)=-Bmn(m,n,i)*pcurr(i)*(m/pi + n*fac2/pi)
     $       *(m*iota(i) - n)/((m*iota(i) - n)**2 + (5.d-3*m)**2)
750      continue
         do 751 n=1,nvf
            Xmn(0,n,i)=-Bmn(0,n,i)*pcurr(i)*( - fac2/pi)
            Xmn(0,-n,i)=-Bmn(0,-n,i)*pcurr(i)*( - fac2/pi)
 751     continue
         Xmn(0,0,i) = 0.

c        calculate surface integrals

         av1(i) = 0.
         av2(i) = 0.
         av3(i) = 0.
         av4(i) =0.
         facdel = hu*hv
         nufm = min(nuf*.5,12.0)      !  normally  nuf*.5
         nvfm = min(nvf*.5,12.0)      !  normally  nvf*.5
         do 7500 j=1,nu
         do 7500 k=1,nv
            sqgjk = 0.
            gsjk  = 0.
            xjk   = 0.
            do 7400 m = 0,nufm
            do 7400 n = -nvfm,nvfm
c            do 7400 m = 0,nuf
c            do 7400 n = -nvf,nvf
               arg = pi2*(m*(j-1)*hu-n*(k-1)*hv)
               cs = (cos(arg))*(1.-m**2/nufm**2)*(1.-n**2/nvfm**2)
c               cs = cos(arg)
               sqgjk = sqgjk + Bmn(m,n,i)*cs
               gsjk  = gsjk +grads2b(m,n)*cs
               xjk   = xjk + Xmn(m,n,i)*cs
7400        continue
            temjk = (sqgjk/gsjk)**2
            av1(i) = av1(i) + temjk*facdel
            av2(i) = av2(i) + sqgjk**2*facdel
            av3(i) = av3(i) + xjk*temjk*facdel
            av4(i) = av4(i)+xjk**2*temjk*facdel
7500     continue
 1500 continue

      m0out = nuf
      n0out = nvf
      do 1503 m=0,nuf
         do 1503 n=-nvf,nvf
            do 1503 i=1,ns-1
               BMODout(m,n,i) = BMODmn(m,n,i)
 1503 continue

      do 2498 i=0,ns-1
 2498 continue
      bnorm = 1.
      endif

c     Second Mercier loop: quantities evaluated at mesh points

      if (cbmn) then

         call deliota

c        Tabulate Mercier quantities

         do 2000 i = 1,ns-1
            ss1=s(i-1,1)
            ss2=s(i,1)
            ssm=s(i,0)
            dels=ss2-ss1
            tcurri  = 0.5*(curr(i-1) + curr(i))
            tcurrip = (curr(i) - curr(i-1))/dels
            pcurri  = 0.5*(pcurr(i-1)+ pcurr(i))
            pcurrip = (pcurr(i) - pcurr(i-1))/dels
            aiotai  = 0.5*(iota(i) + iota(i-1))
            aiotaip = (iota(i) - iota(i-1))/dels
            ftfpij  = -pi*(pcurri + aiotai*tcurri)
            ppi     = 0.5*(pp(i) + pp(i-1))
            vppi    = (volp(i) - volp(i-1))/dels
            av1i    = 0.5*(av1(i) + av1(i-1))
            av2i    = 0.5*(av2(i) + av2(i-1))
            av3i    = -0.5*(av3(i) + av3(i-1))*ppi/ftfpij
            av4i    = 0.5*(av4(i) + av4(i-1))*(ppi/ftfpij)**2

            X00     = (tcurrip*pcurri - pcurrip*tcurri)/ftfpij
            ak12    = tcurrip*pi*aiotaip
            ak13    = vppi*ppi
            ak14    = ppi**2*av2i/ftfpij
            ak15    = (X00**2*av1i + 2.*X00*av3i + av4i)*ftfpij
            ak21    = -0.5*pi**2*aiotaip
            ak22    = (av3i + X00*av1i)*ftfpij
            ak3     = av1i*ftfpij
            amer    = (ak21 + ak22)**2 
            amer    = amer - ak3*(ak12+ak13+ak14+ak15)
            dw      = -ak3*(ak13 + ak14)*ssm**2
            dg      = (ak22**2 - ak3*ak15)*ssm**2
            di      = amer*ssm**2
            omega(i) = di/(irot+aiotai)**2/pi**6
 1900 format(1p4e12.4)
 2000    continue
      endif
c          End second Mercier Loop

      return
      end

      subroutine deliota
c
c     Estimate the change in iota due to a bootstrap current.
c
c
      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2
c
      parameter(mx=160,my=160)
      parameter(ntheta=100,nphi=100)
c     dimension ainv(2,2),cmat(2)
      dimension psiarr(0:nsd),b0arr(0:nsd),cint(0:nsd)
      dimension rhoij(0:ntheta,0:nphi),
     .     rhomn(0:nufd,-nvfd:nvfd),gij(0:ntheta,0:nphi),
     .     gmn(0:nufd,-nvfd:nvfd)
      common /deltacomm/ Pars2(0:nsd),
     . deltamn(0:nufd,-nvfd:nvfd,nsd),Parsmn(0:nufd,-nvfd:nvfd,nsd)

c     Set up constants

      dtheta=1./float(ntheta+1)
      dphi  =1./float(nphi+1)
      nsurf=ns-1
      nper = Q
      m0 = min(nuf*.5,4.0)     ! normally  4
      n0 = min(nvf*.5,12.0)     ! normally  nvf*.5

      do 2 i=1,nsurf
         psiarr(i) = s(i,1)                   !s values at centers
         b0arr(i)  = BMODmn(0,0,i)  
 2    continue

c     Find field maxima

      xmin = -1./2.
      xmax =  -xmin
      ymin =-0.5
      ymax = 0.5
      delx = (xmax - xmin)/float(mx-1)
      dely = (ymax - ymin)/float(my-1)
      dels = 1./nsurf
      do 100 k = 1,nsurf
         psi = psiarr(k)
         aiota = iota(k)
         aicur = curr(k)
         gcur =  pcurr(k)
         gfac =  gcur + aiota*aicur

c     set up z array

         bmax = 0.
         bmin = 1000.
         do 9 j = 1,my
            theta = ymin + dely*(j - 1.)
            do 9 ii = 1,mx
               phi = xmin + delx*(ii - 1.)
               bmod = 0.
               do 8  m=0,m0
                  do 8  n=-n0,n0
                     arg = pi2*(m*theta - n*phi)
                     co = cos(arg)
                     bmod = BMODmn(m,n,k)*co + bmod
8              continue
               if (bmod.gt.bmax) then
                  bmax = bmod
                  themax = theta
                  phimax = phi
               endif
               if (bmod.lt.bmin) then
                  bmin = bmod
                  themin = theta
                  phimin = phi
               endif
9        continue

         thetao = themax
         phio   =   phimax
         bmod   =   bmax           !
         b0 = bmod

c     Accumulate rhoij=v_parallel/b
c          (over one field period)

c        amu = 1./b0
         amu = .75/b0
         do 26 i=0,ntheta
         do 26 j=0,nphi
            rhoij(i,j) = 0.
            bmod=0.
            dbdt = 0.
            dbdp = 0.
            do 24 m=0,m0
             do 24 n=-n0,n0
               arg = pi2*(m*i*dtheta - n*j*dphi)
               co = cos(arg)
               si = sin(arg)
               bmod = bmod + BMODmn(m,n,k)*co
 24         continue
            bound = 1. - amu*bmod
            eps = 1.d-4        !
            if (bound.lt.eps) then
               rhoij(i,j) = 0.
            else
               vpara = sqrt(2.*(1. - amu*bmod))
               rhoij(i,j) = vpara/sqrt(bmod**2 + 1.d-4)*b0/sqrt(2.)  !
            endif
            gij(i,j) = 1./(bmod**2 + 1.d-4)
 26      continue

c     Fourier transform rhoij and gij

         do 28 m=0,m0
          do 28 n=-n0,n0
            rhomn(m,n) = 0.
            gmn(m,n) = 0.
            do 27 i=0,ntheta
               do 27 j=0,nphi
                  arg = pi2*(m*i*dtheta - n*j*dphi)
                  rhomn(m,n)=rhomn(m,n) + rhoij(i,j)*cos(arg)
                  gmn(m,n)  =gmn(m,n)   + gij(i,j)*cos(arg)
 27        continue
            rhomn(m,n) = rhomn(m,n)*dtheta*dphi*2.
            gmn(m,n)   = gmn(m,n)*dtheta*dphi*2.
 28     continue
         do 29 n=-n0,n0
            rhomn(0,n) = rhomn(0,n)/2.
            gmn(0,n)   = gmn(0,n)/2.
 29     continue

c        Accumulate deltamn

         do 30 m=0,m0
          do 30 n=-n0,n0
            if (m.eq.0.and.n.eq.0) goto 30
             deltamn(m,n,k) =(m*gcur + n*aicur)*rhomn(m,n)
     .       *(m*aiota-n)/((m*aiota-n)**2+(5.d-3*m)**2)/gfac    !
  30     continue
c        Find (0,0) coefficient, a very important choice!
         deltamn(0,0,k) = 0.
         temp = 0.
         do 40 m=0,m0
          do 40 n=-n0,n0
            arg = pi2*(m*themax - n*phimax)
            temp = temp - deltamn(m,n,k)*cos(arg)
  40     continue
         deltamn(0,0,k) = temp         !  HJG vs PRG, QAS vs QHS
c        deltamn(0,0,k) = 0.0          !  HJG vs PRG, QAS vs QHS

c        Sum the Parseval series of coefficients (which is
c        equivalent to evaluating the surface averages).

         Pars1=0.
         do 45 m=1,m0
            do 45 n=-n0,n0
               Parsmn(m,n,k) = 0.5*deltamn(m,n,k)*rhomn(m,n)
               Pars1 = Pars1 + Parsmn(m,n,k)
 45      continue
         do 46 n=-n0,n0
            Parsmn(0,n,k) = deltamn(0,n,k)*rhomn(0,n)
            Pars1 = Pars1 + Parsmn(0,n,k)
 46      continue
         Pars2(k) = rhomn(0,0)
         deltapb    = 1.5*Pars1/Pars2(k)/nper
         del0arr(k) = deltapb
100   continue

c     Helical radius for aspect ratio

      Rhel = nper*(pcurr(nsurf)-irot*curr(nsurf))
     .        /pi2/BMODmn(0,0,nsurf)

c     Boostrap current at centers

      av_beta = avp/sqrt(avb**2 + 1.d-4)     !
      del0arr(0) = 0.
      psiarr(0)  = 0.
      b0arr(0)   = b0arr(1)
      del0arr(nsurf+1) = del0arr(nsurf)
      psiarr(nsurf+1)  = 1.
      b0arr(nsurf+1)   = b0arr(nsurf)


      aint = 0.
      avb0 = 0.
      cint(0) = 0.
      do 310 i=1,nsurf
         smid  = 0.5*(psiarr(i) + psiarr(i-1))
         dels  = psiarr(i) - psiarr(i-1)
         delval= 0.5*(del0arr(i) + del0arr(i-1))
         pcurr_mid = 0.5*(pcurr(i) + pcurr(i-1))

         cint(i) = cint(i-1) + 
     .            2.37*av_beta*delval*(1.-smid)*pcurr_mid*dels

         aint  = aint + delval*dels*(1.-smid)

         b0mid = 0.5*(b0arr(i) + b0arr(i-1))
         avb0  = avb0 + b0mid*dels 

 310  continue


      do 311 i=0,nsurf
         curr0(i) = cint(i)
 311  continue


      asp_hel = Rhel*sqrt(avb0)
      delta_iota = aint*2.37*asp_hel**2/nper

      return
      end



      subroutine coord
c     compute coordinate system x0,y0,x2,y2,H and v derivatives      
c     from the Fourier series for x0 and y0
c
c
      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2
c
      real*8 filF(0:nvf2d)
      call FTV(-1,x0,x0F(0,tnew))
      call FTV(-1,y0,y0F(0,tnew))
      call FTDV(-1,x0v,x0F(0,tnew))
      call FTDV(-1,y0v,y0F(0,tnew))

c     filtered values x0filv of x0v for use in axis residue
c     The filter starts at the n=Qrun+1 term.  (Where Qrun=number of
c     field periods for the calculation)
      do 100 kf=0,nvf2
         n=mod(kf,nvf+1)
         x=(pi*(n-Qrun))/max(1,(nvfAX-Qrun))
         if (n.le.Qrun) filF(kf)=x0F(kf,tnew)
         if (n.gt.Qrun) filF(kf)=x0F(kf,tnew)*sin(x)/x
         if (n.gt.nvfAX) filF(kf)=0
100   continue
      call FTDV(-1,x0filv,filF)

c     filtered values y0filv of y0v for use in axis residue
      do 110 kf=0,nvf2
         n=mod(kf,nvf+1)
         x=(pi*(n-Qrun))/max(1,(nvfAX-Qrun))
         if (n.le.Qrun) filF(kf)=y0F(kf,tnew)
         if (n.gt.Qrun) filF(kf)=y0F(kf,tnew)*sin(x)/x
         if (n.gt.nvfAX) filF(kf)=0
110   continue
      call FTDV(-1,y0filv,filF)

      do 190 j=1,nu
      do 190 k=1,nv
         x2(k,j) = x1(k,j)-x0(k)    
         y2(k,j) = y1(k,j)-y0(k)    
         x2v(k,j)= x1v(k,j)-x0v(k)  
         y2v(k,j)= y1v(k,j)-y0v(k)  
         x2filv(k,j)= x1v(k,j)-x0filv(k)  
         y2filv(k,j)= y1v(k,j)-y0filv(k)  
         x2u(k,j)= x1u(k,j)
         y2u(k,j)= y1u(k,j)
         rscale(k,j)=x2(k,j)**2 + y2(k,j)**2
         rwall(k,j) = 1
         rwallu(k,j)= 0
         rwallv(k,j)= 0
         H(k,j)=y2u(k,j)*x2(k,j) - x2u(k,j)*y2(k,j)
190   continue
      end



      subroutine inhmode
c     initialize modes for inhomogeneous problem in stability test
c     mode from Mag. Equ. & Stability of Stellarators, page 19.
c     mode number:  cos(pi2*(mh*u -  nh*v))
c     strength: lamR(i),lamPSI(i) and lamAX 
c     inhPDE is set .true. if there is a mode for the R and PSI equations.
c     inhAX is set .true. if there is a mode for the AXIS equation.
c     The inhomogeneous term will be added for iterations between
c     starth and stoph.  
c
c
      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2
c
      real*8 sn,sc,facR,facPSI,u,v
      inhPDE=.false.
      inhAX=.false.
      if ((mh.eq.0) .and. (nh.eq.0)) return
      facR=mode
      facPSI=mode
      lamAX=2*mode/5.0

c     R and PSI inhomogeneous term
      inhPDE=.true.
      do 600 i=0,ns-1
         sn=s(i,0)
         sc=s(i,1)
         lamR(i)=sn**2 * (1-sn) * facR
         lamPSI(i)=-sc*(1-2*sc) * facPSI
600   continue
      do 610 j=1,nu
      do 610 k=1,nv
         u=(j-1)*hu
         v=(k-1)*hv
         tmp1(k,j)=cos(pi2*(mh*u - nh*v))                          !
         tmp2(k,j)=sin(pi2*(mh*u - nh*v))                          !
c        tmp1(k,j)=sin(pi2*(mh*u - nh*v))                          !
c        tmp2(k,j)=cos(pi2*(mh*u - nh*v))                          !
610   continue
      call FTV(2,tmp1,ftwork)
      call FTU(2,ftwork,LRFh)
      call FTV(2,tmp2,ftwork)
      call FTU(2,ftwork,LPSIFh)

c     axis inhomogeneous term
      if (mh.eq.1) then 
         inhAX=.true.
         do 650 k=1,nv
            v=(k-1)*hv
            tmp1(k,1)=cos(pi2*(nh+irot)*v)                         !
            tmp2(k,1)=sin(pi2*(nh+irot)*v)                         !
c           tmp1(k,1)=sin(pi2*(nh+irot)*v)       !                 !
c           tmp2(k,1)=cos(pi2*(nh+irot)*v)       !                 !
650      continue
         call FTV(1,tmp1,LXFh)
         call FTV(1,tmp2,LYFh)
      endif
      end



      subroutine extend
c     extend solution to Qext field periods for stability test
c
c
      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2
c
      integer oldk,newk,oldnvf

c     initialize constants      
      Qrun=Qext
c     convert nh back to irot=0 coordinate system
      nh=nh+irot*mh
      irot=Qext*irot
c     convert nh to new irot coordinate system
      nh=nh-irot*mh
      nv=Qext*nv
      hv=1d0/nv
      oldnvf=nvf
      nvf=Qext*nvf
      nvfAX=Qext*nvfAX
      nvf2=2*nvf+1
      zfac=Qext*zfac
      pd1=pd1/Qext**2
      pd2=pd2/Qext**2
c     recompute Fourier transform matrices
      call FTstore
c     recompute outer wall
c     outer wall quantities from del(m,n) 
      do 50 j=1,nu
      do 50 k=1,nv
         u=(j-1)*hu
         v=(k-1)*hv
50       call wall
     $   (u,v,x1(k,j),y1(k,j),x1u(k,j),y1u(k,j),x1v(k,j),y1v(k,j))
      
c     extend flux
      do 450 i=0,ns-1
450      iota(i)=Qext*iota(i)

c     extend axis, R, PSI at both time levels
      do 100 tuse=0,1
      do 100 k2=1,0,-1
      do 100 kf=nvf,0,-1
         newk=k2*(nvf+1)+kf
         oldk=k2*(oldnvf+1)+kf/Qext
         if (mod(kf,Qext).eq.0) then
            x0F(newk,tuse)=x0F(oldk,tuse)
            y0F(newk,tuse)=y0F(oldk,tuse)
            do 90 i=0,ns
            do 90 jf=0,nuf2
               RF(newk,jf,i,tuse)=RF(oldk,jf,i,tuse)
               PSIF(newk,jf,i,tuse)=PSIF(oldk,jf,i,tuse)
90          continue
         else
            x0F(newk,tuse)=0
            y0F(newk,tuse)=0
            do 95 i=0,ns
            do 95 jf=0,nuf2
               RF(newk,jf,i,tuse)=0
               PSIF(newk,jf,i,tuse)=0
95          continue
         endif
100   continue
      end



      subroutine output
c     organize and print output
c
      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2

      integer skip
      real*8 mx,coeff(0:nsd)

         open(unit=10,file='spectrum',form='formatted')

c     do not print more than 7 numbers across the page
      skip=ns/8 + 1
      ptol=5.d-3


      if (nprint.ge.10) then
c        norms of cos(m*u - n*v) terms in R
      write(7,*) char(12)
         write(7,*) '                     R/sqrt(s)'
         write(7,*)
         write(7,*) '                            s'
         write(7,1010) 'm','n',(s(i,0),i=1,ns-1,skip)
         write(7,*)
         do 100 jf=0,nuf
         do 100 kf2=-nvf,nvf
            kf=abs(kf2)
            mx=0
            do 90 i=1,ns-1
               coeff(i)=.5*(RF(kf,jf,i,tnew)+                      !
     $         sign(1,kf2)*RF(kf+nvf+1,jf+nuf+1,i,tnew))           !
c              coeff(i)=.5*(RF(kf,jf+nuf+1,i,tnew)-                !
c    $         sign(1,kf2)*RF(kf+nvf+1,jf,i,tnew))                 !
               coeff(i)=coeff(i)/srt(i,0)
c              combine some of the cos(m*u+n*v) with the cos(m*u-n*v) terms
               if (jf.eq.0 .or. kf.eq.0) coeff(i)=coeff(i)*2
               if (jf.eq.0 .and. kf2.lt.0) coeff(i)=0
               mx=max(mx,abs(coeff(i)))
90          continue
            if (mx.ge.ptol)
     $      write(7,1000) jf,kf2+irot*jf,(coeff(i),i=1,ns-1,skip)
100      continue      
      endif


c     avb = integral .5*B*B,  thus <B*B> = 2*avb/vol
c     beta = integral p / integral .5*B*B = avp/avb
      BL2=2*avb/sqrt(vol**2 + 1.d-4)
      av_beta = avp/sqrt(avb**2 + 1.d-4)     !
      write(7,*) char(12)
      write(7,1050) ' beta  = ',av_beta
      write(7,1050) '<B*B>  = ',BL2
      write(7,*)
      write(7,*)
      write(7,*)

c     Ip/L = pcurr(i)/zfac no longer printed
      write(7,1020) 's',"V''",'Omega',"p'",'I/s','iota',
     $    'deliota','n','x0','y0'
      write(7,*)
      do 400 i=0,max(ns-1,nvfAX)
c        magnetic well = vol'' / vol' 
         if (i.ge.1 .and. i .le.ns-1)
     $   well=(volp(i)-volp(i-1))/(hs(i,0)*.5*(volp(i)+volp(i-1)))
         piota=iota(i) + irot
            if (i.ge.1 .and. i.lt.ns-1)
c    $del0arr(i)=0.75*(1-s(i,1)**2)*av_beta*del0arr(i)/(Q*ep**2)       !
     $del0arr(i)=5.d-1*av_beta*del0arr(i)/(Q*ep**2)       !
         if (i.gt.nvfAX) then
            if (i.lt.ns-1) then
            write(7,1021) s(i,1),well,0.5*(omega(i)+omega(i+1)),
     $      ppout(i),curr(i)/s(i,1),piota,del0arr(i)
            else
            write(7,1024) s(i,1),well,ppout(i),
     $      curr(i)/s(i,1),piota
            endif
         else if (i.gt.ns-1) then
            write(7,1022) i,x0F(i,tnew),y0F(i+nvf+1,tnew)
         else if (i.eq.0) then
            write(7,1023) s(i,1)
     $      ,ppout(i),curr(i)/s(i,1),
     $      piota,i,x0F(i,tnew),y0F(i+nvf+1,tnew)
         else
            if (i.lt.ns-1) then
            write(7,1021) s(i,1),well,0.5*(omega(i)+omega(i+1)),
     $      ppout(i),curr(i)/s(i,1),
     $      piota,del0arr(i),
     $      i,x0F(i,tnew),y0F(i+nvf+1,tnew)
            else
            write(7,1024) s(i,1),well,ppout(i),curr(i)/s(i,1),
     $      piota,
     $      i,x0F(i,tnew),y0F(i+nvf+1,tnew)
            endif
         endif
400   continue


      ptolBmn=0.001
      if (nprint.ge.12) then
         write(7,*)
         write(7,*) '                     Bmn'
         write(7,*)
         write(7,*) '                            s'
         write(7,1010) 'm','n',(s(i,1),i=1,ns-1,skip)
         write(7,*)
         do 405 m=0,nuf
            do 405 n=-nvf,nvf
               do 405 i=1,ns-1
                   Bmn(m,n,i) = -Bmn(m,n,i)
     $             /(pi*(pcurr(i)+iota(i)*curr(i)))
 405     continue
         do 50 m=0,nuf
         do 50 n=-nvf,nvf
c                Bmn(m,n,i) is the cos(mu-nv) Fourier coefficient
c                of 1/B**2  at the ith flux surface.

            nnn=abs(n+irot*m)
            if ((m.ge.nuf/2).or.(nnn.ge.nvf/2)) go to 50
               mx=0
               do 40 i=1,ns-1
40                mx=max(mx,abs(Bmn(m,n,i)))
                  if (mx.ge.ptolBmn)  write(7,1000) m,n+irot*m,
     .                     (Bmn(m,n,i),i=1,ns-1,skip)
50       continue      

      endif

      mf=4
      nf=4
         do 505 m=0,mf
            do 505 n=-nf,nf
               do 505 i=1,ns-1
  505 write(10,2000) Bmn(m,n-irot*m,i)

1000  format(2i4,99f9.3)
1010  format(2a4,99f9.3)
1020  format(7a9,1x,a3,a6,1x,a6)
1021  format(7f9.4,1x,i3,2f7.3,2e15.5)
1024  format(2f9.4,9x,3f9.4,10x,i3,2f7.3,2e15.5)
1023  format(f9.4,18x,3f9.4,10x,i3,2f7.3,2e15.5)
1022  format(54x,10x,i3,2f7.3,2e15.5)
1030  format(a4,99a10)
1031  format(i4,5f10.4)
1050  format(a10,f6.4)
2000  format(f9.3)
      end




      subroutine restart(iflag)
      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2
c
c     iflag=0  save data in file "save"
c     iflag=1  read in data from file "restart"
      if (iflag.eq.0) then
         open(unit=28,file='save',form='unformatted')
         write(28)
     $        (((  RF(kf,jf,i,1),kf=0,nvf2),jf=0,nuf2),i=0,ns)
         write(28)
     $        (((PSIF(kf,jf,i,1),kf=0,nvf2),jf=0,nuf2),i=0,ns)
         write(28) (x0F(kf,1),kf=0,nvf2)
         write(28) (y0F(kf,1),kf=0,nvf2)
         write(28) (iota(i),i=0,ns)
      endif
      if (iflag.eq.1) then
         open(unit=28,file='restart',form='unformatted')
         read(28)
     $        (((  RF(kf,jf,i,1),kf=0,nvf2),jf=0,nuf2),i=0,ns)
         read(28)
     $        (((PSIF(kf,jf,i,1),kf=0,nvf2),jf=0,nuf2),i=0,ns)
         read(28) (x0F(kf,1),kf=0,nvf2)
         read(28) (y0F(kf,1),kf=0,nvf2)
         if (stell) read(28) (iota(i),i=0,ns)
      endif
c      close(28)
      end


      subroutine plot

c     Modified by GBM 10/6/07 to write out data for subsequent plot
c     print out data in ascii (not binary)

c     data is written to unit 21

      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2

c
      parameter (numsfd=500)
      parameter (nsdd=2*nsd)
      integer step
c     NCAR uses real*4 variables
      real*4 xp2(0:numpd),yp2(0:numpd),w,x0i,y0i
      real*4 xp3(0:numsfd,0:nsdd),yp3(0:numsfd,0:nsdd)
      real*4 xstart,ystart,bd,axmax,mx
      real*4 errpp(numpd,10)
      real*8 cosu(0:numsfd),sinu(0:numsfd),cosv(0:nvfd),sinv(0:nvfd)
      real*8 x1i(0:numsfd),y1i(0:numsfd)

c     for plotting the flux surfaces, evaluate the Fourier series at
c     at least 72 points for a smooth curve.  A multiple of nu points
c     in u is used so that the rays can be easily plotted
c     and they will be at the true mesh points (assuming the v=const
c     cross section is on an exact v mesh point which is only the case
c     when nv is a multple of 4).

      write(21,997) nu

      numray=nu
      numsf=(72/numray + 1 ) * numray
      if (numsf.gt.numsfd) then
          write(7,*) 'numsf > numsfd, no plots produced'
          return
      endif

      write(21,997) nump
      do 17 jj=1,nump
         write(21,999) (errp(jj,kk),kk=1,5)
 17   continue

      axmax=0

c     maximum value scaling plots
      do 100 j=1,nu
      do 100 k=1,nv
      axmax1=axmax
100      axmax=max(abs(x1(k,j)),abs(y1(k,j)),axmax1)

      write(21,999) axmax

c     store some trig functions
      do 110 j2=0,numsf
         cosu(j2)=cos((pi2*j2)/numsf)
110      sinu(j2)=sin((pi2*j2)/numsf)

c     plot 4 Poincare sections of the torus
      do 400 kp=1,1           !!??!!     normally  0,3
      v=kp*.25
      xstart=mod(kp,1)*1.0      !!??!!   normally .5
      ystart=(1-(kp/1))*1.0      !!??!!   normally .5
      bd=.30

c     some trig functions of v
      do 95 kf=0,nvf
         cosv(kf)=cos(pi2*kf*v)
95       sinv(kf)=sin(pi2*kf*v)

c     compute axis in this Poincare section
      x0i=0
      y0i=0
      do 96 kf=0,nvfAX
         x0i=x0i + x0F(kf,tnew)*cosv(kf) + x0F(kf+nvf+1,tnew)*sinv(kf)
96       y0i=y0i + y0F(kf,tnew)*cosv(kf) + y0F(kf+nvf+1,tnew)*sinv(kf)

c     compute outer wall in this Poincare section as a function of u
      do 210 j2=0,numsf
         u=real(j2)/numsf
         call wall(u,v,x1i(j2),y1i(j2),x1iu,y1iu,x1iv,y1iv)
210   continue

c     compute x and y in this Poincare section as functions of u and s
      step=max(1,(ns-1)/22)     !  normally  /7)

      write(21,997) ns,step,numsf

      do 230 i=ns,0,-step
         do 220 j2=0,numsf
            Ri=0
            rwi=1
            x2i=(x1i(j2)-x0i)/rwi
            y2i=(y1i(j2)-y0i)/rwi
            if (i.eq.ns) then
               Ri=rwi
            else
               do 200 jf=0,nuf
               do 200 kf=0,nvf
                  jfi=mod(jf*j2,numsf)
                  Ri=Ri +    RF(kf,jf,i,tnew)*cosu(jfi)*cosv(kf)
     $       +         RF(kf,jf+nuf+1,i,tnew)*sinu(jfi)*cosv(kf)
     $       +         RF(kf+nvf+1,jf,i,tnew)*cosu(jfi)*sinv(kf)
     $          +RF(kf+nvf+1,jf+nuf+1,i,tnew)*sinu(jfi)*sinv(kf)
200            continue
            endif
            xp3(j2,i)=x0i + Ri * x2i
            yp3(j2,i)=y0i + Ri * y2i
220      continue
         write(21,999) (xp3(j2,i),yp3(j2,i),j2=0,numsf)
230   continue
c     plot axis

      write(21,999) x0i,y0i

400   continue

      write(21,997) nprint,nump,pmod

 997  format(4i8)
 999  format(6e12.4)

      return
      end

      subroutine norm(mx,L2sq,imax,jmax,f,ld,n1,n2)
c     compute maximum norm and L2 norm of 2D array f.
c     f=f(i,j) where i=0..n1, j=0..n2, and f is dimensioned by f(0:ld,0:*)
c
      implicit real*8 (a-h,o-z)
      real*8 f(0:ld,0:*),x,mx,L2sq
      mx=0
      imax=-1
      jmax=-1
      L2sq=0
      do 100 j=0,n2
      do 100 i=0,n1
         x=abs(f(i,j))
         L2sq=L2sq+x**2
         if (x.gt.mx) then
            mx=x
            imax=i
            jmax=j
         endif
100   continue
      L2sq=L2sq/((n1+1)*(n2+1))
      end



      subroutine tri(a,b,c,u,kf,jf,n1,n2,n3)
c     solves the tridiagonal system Au=y in the s variable (i=1..ns-1) 
c     at each point in Fourier space (kf=0..nvf2,jf=0..nuf2).
c     y is originaly stored in u, and is overwritten by the solution.
c     A is the tridiagonal matrix with b(i) on the diagonal,
c     a(i) on the lower diagonal and c(i) on the upper diagonal.
c     a(1) and c(ns-1) are not defined.  
c     n1,n2 and n3 must be equal to nvf2d,nuf2d and nsd respectivly.
c     (Sun f77 requires dummy arrays to have adjustable diminsions)
c     from Press, Flannery, Teukolsky, Vetterling, "Numerical Recipes"
c
c
      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2
c
      integer jf,kf,i,n1,n2,n3,n
      real*8 work(nsd),a(n3),b(n3),c(n3),u(0:n1,0:n2,0:n3),pivot

      n=ns-1
      if(b(1).eq.0.) stop 'trid error 1'
      pivot=b(1)
      u(kf,jf,1)=u(kf,jf,1)/pivot
      do 11 i=2,n
        work(i)=c(i-1)/pivot
        pivot=b(i)-a(i)*work(i)
        if(pivot.eq.0.) stop 'trid error 2'
        u(kf,jf,i)=(u(kf,jf,i)-a(i)*u(kf,jf,i-1))/pivot
11    continue
      do 12 i=n-1,1,-1
        u(kf,jf,i)=u(kf,jf,i)-work(i+1)*u(kf,jf,i+1)
12    continue
      return
      end



      subroutine FTstore
c     Store matrices of trig functions for Fourier transforms
c     this subroutine also contains entry points for the 
c     Fourier transform routines FTV,FTU,FTDV,FTDU,PDU and PDV
c     Fourier transforms are performed with the BLAS3 general matrix
c     multiply routine SGEMM on the CRAYs.  Use DGEMM for machines where
c     real*8 is double precision instead of single precision.  
c     For other machines, the fastest
c     available matrix multiply routine should be used.  
c     The FFT is not used because breakeven point between FFT and matrix
c     multiply is larger than the matrix sizes we normaly use.
c
c
      implicit real*8 (a-h,o-z)
      integer Q1
c     The user should only change the first parameter statement:
c
      parameter (ns1=12,nuf1=21,nvf1=21,Q1=1)
c
      parameter (nu1=3*nuf1,nv1=3*nvf1)
      parameter (nvd=nv1*Q1,nvfd=nvf1*Q1)
      parameter (nud=nu1,nufd=nuf1,nsd=ns1)
      parameter (nuf2d=2*nufd+1,nvf2d=2*nvfd+1)
      parameter (numpd=1200)
      parameter (one1=1.,zero0=0.)

      integer stoph,starth,Q,Qext,Qrun,tnew,told,tuse,pmod,bootit
      common /globali/
     $ ns,nu,nv,nuf,nvfAX,nvf,nuf2,nvf2,numit,numext,ic,nprint,
     $ mh,nh,stoph,starth,locPSI,locPSIn,locPSIm,
     $ locR,locRm,locRn,locAXn,icont,imerc,
     $ Q,Qext,Qrun,it,irot,tnew,told,tuse,nav,nump,pmod
     $ ,nub,nvb,nufb,nvfb,nboot,bootit,nfinal
      logical negjac,inhAX,inhPDE,inhON,stell,rays
      common /flags/negjac,inhAX,inhPDE,inhON,stell,rays
c
      real*8 JD,JFAC,jac,iota,iota1,iota0,LPSIF,LRF,
     $ lamR,lamPSI,lamAX,LRFh,LPSIFh,LXFh,LYFh,LXF,LYF,LX,LY,
     $ M1ss,M1uu,M1vv,M2uu,M2vv,M3,lthick,modeval,mode,model2
      common /globalr0/
     $ del(-nufd:nufd,-nvfd:nvfd),lthick,
     $ zone(-nufd:nufd,-nvfd:nvfd),fixAX,
     $ s(0:nsd,0:1),srt(0:nsd,0:1),bsq(nvd,nud,0:1),
     $ M1ss(0:nsd),M1uu(0:nsd),M1vv(0:nsd),
     $ M2uu(0:nsd),M2vv(0:nsd),M3(0:nsd),
     $ alpha,avb,avp,vol,volp(0:nsd),omega(0:nsd),delta_iota,
     $ hs(0:nsd,0:1),rscale(nvd,nud),asp_hel,
     $ TuTu(nvd,nud),TuTv(nvd,nud),TvTv(nvd,nud),
     $ TuTs(nvd,nud),TvTs(nvd,nud),
     $ rwall(nvd,nud),rwallu(nvd,nud),rwallv(nvd,nud),
     $ pivots(0:nsd),errp(numpd,10),Bmn(0:nufd,-nvfd:nvfd,nsd),
     $ BMODmn(0:nufd,-nvfd:nvfd,nsd),del0arr(0:nsd),
     $ Xmn(0:nufd,-nvfd:nvfd,0:nsd),
     $ pd1,pd2,pd3,artvR,artvPSI,artvAX,LXF(0:nvf2d),LYF(0:nvf2d),
     $ LX(nvd),LY(nvd),modeval,model2,mode
      common /globalr1/
     $ hsig,hu,hv,ht,aR,aPSI,aAX,eiota,eR,ePSI,eAX,safi,emin,
     $ errI,errPSI,errR,errAX,errPSI2,errR2,errAX2,
     $ jac,ener,errl2R,errl2PSI,errl2AX,
     $ RF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ R(nvd,nud,0:1),Rs(nvd,nud,0:1),
     $ Ru(nvd,nud,0:1),Rv(nvd,nud,0:1),
     $ PSIF(0:nvf2d,0:nuf2d,0:nsd,0:1),
     $ PSIu(nvd,nud,0:1),PSIv(nvd,nud,0:1),
     $ x0(nvd),y0(nvd),x0v(nvd),y0v(nvd),
     $ x0filv(nvd),y0filv(nvd),x2filv(nvd,nud),y2filv(nvd,nud),
     $ x0F(0:nvf2d,0:1),y0F(0:nvf2d,0:1),
     $ x1(nvd,nud),y1(nvd,nud),
     $ x1u(nvd,nud),x1v(nvd,nud),
     $ y1u(nvd,nud),y1v(nvd,nud),
     $ x2(nvd,nud),y2(nvd,nud),
     $ x2v(nvd,nud),y2v(nvd,nud),
     $ x2u(nvd,nud),y2u(nvd,nud)
      common /globalr2/
     $ LPSIF(0:nvf2d,0:nuf2d,0:nsd),LRF(0:nvf2d,0:nuf2d,0:nsd),
     $ D(nvd,nud,0:1),H(nvd,nud),
     $ avxu(nvd,nud),avxv(nvd,nud),avyu(nvd,nud),
     $ avyv(nvd,nud),xs(nvd,nud),ys(nvd,nud),
     $ JD(nvd,nud),JFAC(nvd,nud),
     $ BS(nvd,nud,0:1),BV(nvd,nud,0:1),BU(nvd,nud,0:1),
     $ tmp1(nvd,nud),tmp2(nvd,nud),
     $ ftwork(0:nvf2d,nud),
     $ pcurr(0:nsd),curr(0:nsd),curr0(0:nsd),
     $ iota(0:nsd),iota1,iota0,
     $ p(0:nsd),ppout(0:nsd),pp(0:nsd),p0,xpr,zpr,ypr,
     $ ep,bndj,zlen,zfac,
     $ lamR(0:nsd),lamPSI(0:nsd),lamAX,LRFh(0:nvf2d,0:nuf2d),
     $ LPSIFh(0:nvf2d,0:nuf2d),LXFh(0:nvf2d),LYFh(0:nvf2d),
     $ pi,pi2
c
      real*8 tmp,f,fhat,
     $ FM(nvd,0:nvf2d),
     $ DFM(nvd,0:nvf2d),
     $ IFM(0:nvf2d,nvd),
     $ DIFM(0:nvf2d,nvd),
     $ GM(nud,0:nuf2d),
     $ DGM(nud,0:nuf2d),
     $ IGM(0:nuf2d,nud),
     $ DIGM(0:nuf2d,nud),
     $ PDVM(nvd,nvd),
     $ PDUM(nud,nud)
c     force matrices to be saved between calls to FT subroutines
      common /matrix/FM,DFM,IFM,DIFM,GM,DGM,IGM,DIGM,PDVM,PDUM

c     matrices for 2D Fourier transforms in v
      do 100 kf=0,nvf
      do 100 k=1,nv
         ifac=2
         if (kf.eq.0) ifac=1
         tmp=pi2*hv*mod((k-1)*kf,nv)
         FM(k,kf)=ifac*cos(tmp)
         FM(k,kf+nvf+1)=ifac*sin(tmp)
         IFM(kf,k)=cos(tmp)
         IFM(kf+nvf+1,k)=sin(tmp)
         DFM(k,kf)=kf*FM(k,kf+nvf+1)
         DFM(k,kf+nvf+1)=-kf*FM(k,kf)
         DIFM(kf,k)=-kf*IFM(kf+nvf+1,k)
         DIFM(kf+nvf+1,k)=kf*IFM(kf,k)
100   continue

c     matrices 2D for Fourier transform in u
      do 200 jf=0,nuf
      do 200 j=1,nu
         ifac=2
         if (jf.eq.0) ifac=1
         tmp=pi2*hu*mod((j-1)*jf,nu)
         GM(j,jf)=ifac*cos(tmp)
         GM(j,jf+nuf+1)=ifac*sin(tmp)
         IGM(jf,j)=cos(tmp)
         IGM(jf+nuf+1,j)=sin(tmp)
         DGM(j,jf)=jf*GM(j,jf+nuf+1)
         DGM(j,jf+nuf+1)=-jf*GM(j,jf)
         DIGM(jf,j)=-jf*IGM(jf+nuf+1,j)
         DIGM(jf+nuf+1,j)=jf*IGM(jf,j)
200   continue

c     matrix for v differentation via Fourier transform
      do 300 k=1,nv
      do 300 k2=1,nv
         if(k2.eq.k) then
            PDVM(k2,k)=0
         else
            tmp=pi2*hv*mod((k2-k),nv)
            PDVM(k2,k)=
     $                   .5*nvf * sin((nvf+2)*tmp)
     $           - (1.5*nvf+.5) * sin((nvf+1)*tmp)
     $           +  (1.5*nvf+1) * sin(nvf*tmp)
     $           -   .5*(nvf+1) * sin((nvf-1)*tmp) 
            PDVM(k2,k)=PDVM(k2,k)/((1-cos(tmp))**2)
         endif
300   continue

c     matrix for u differentiation via Fourier transform
      do 400 j=1,nu
      do 400 j2=1,nu
         if(j2.eq.j) then
            PDUM(j2,j)=0
         else
            tmp=pi2*hu*mod((j2-j),nu)
            PDUM(j2,j)=
     $                   .5*nuf * sin((nuf+2)*tmp)
     $           - (1.5*nuf+.5) * sin((nuf+1)*tmp)
     $           +  (1.5*nuf+1) * sin(nuf*tmp)
     $           -   .5*(nuf+1) * sin((nuf-1)*tmp) 
            PDUM(j2,j)=PDUM(j2,j)/((1-cos(tmp))**2)
         endif
400   continue
      return



      entry FTV(i,f,fhat)
c     v Fourier transform via a single matrix multiply (SGEMM)
c     i > 0   Fourier transform f -> fhat
c     i < 0   inverse Fourier transform fhat -> f
c     i = 1   f and fhat functions of v only 
c     i = 2   f and fhat functions of both u and v
c
      m=1
      if (abs(i).eq.2) m=nu
      if (i.lt.0) then
         call dgemm('t','n',nv,m,nvf2+1,one1,       ! dgemm ! on SGI,SUN
     $        IFM,nvf2d+1,fhat,nvf2d+1,zero0,f,nvd) 
      else
         call dgemm('t','n',nvf2+1,m,nv,hv,         ! dgemm ! on SGI,SUN
     $        FM,nvd,f,nvd,zero0,fhat,nvf2d+1) 
      endif
      return



      entry FTU(i,f,fhat)
c     u Fourier transform via a single matrix multiply (SGEMM)
      if (i.lt.0) then
         call dgemm('n','n',nvf2+1,nu,nuf2+1,one1,  ! dgemm ! on SGI,SUN
     $   fhat,nvf2d+1,IGM,nuf2d+1,zero0,f,nvf2d+1) 
      else
         call dgemm('n','n',nvf2+1,nuf2+1,nu,hu,    ! dgemm ! on SGI,SUN
     $   f,nvf2d+1,GM,nud,zero0,fhat,nvf2d+1) 
      endif
      return



      entry FTDV(i,f,fhat)
c     v Fourier transform with v derivative 
      m=1
      if (abs(i).eq.2) m=nu
      if (i.lt.0) then
         call dgemm('t','n',nv,m,nvf2+1,pi2,         ! dgemm ! on SGI,SUN
     $        DIFM,nvf2d+1,fhat,nvf2d+1,zero0,f,nvd) 
      else
         call dgemm('t','n',nvf2+1,m,nv,hv*pi2,      ! dgemm ! on SGI,SUN
     $        DFM,nvd,f,nvd,zero0,fhat,nvf2d+1) 
      endif
      return



      entry FTDU(i,f,fhat)
c     u Fourier transform with u derivative
      if (i.lt.0) then
         call dgemm('n','n',nvf2+1,nu,nuf2+1,pi2,    ! dgemm ! on SGI,SUN
     $   fhat,nvf2d+1,DIGM,nuf2d+1,zero0,f,nvf2d+1) 
      else
         call dgemm('n','n',nvf2+1,nuf2+1,nu,hu*pi2,  ! dgemm ! on SGI,SUN
     $   f,nvf2d+1,DGM,nud,zero0,fhat,nvf2d+1) 
      endif
      return



      entry PDV(i,f,fv)
c     v derivative via Fourier transform and inverse transform
      m=1
      if (abs(i).eq.2) m=nu
      call dgemm('t','n',nv,m,nv,pi2*hv,             ! dgemm ! on SGI,SUN
     $     PDVM,nvd,f,nvd,zero0,fv,nvd) 
      return



      entry PDU(i,f,fv)
c     u derivative via Fourier transform and inverse transform
      m=1
      if (abs(i).eq.2) m=nv
      call dgemm('n','n',m,nu,nu,pi2*hu,             ! dgemm ! on SGI,SUN
     $     f,nvd,PDUM,nud,zero0,fv,nvd) 
      return
      end

c     subroutine second(t)  ! Comment this routine on a CRAY
c     returns the CPU time in seconds
c     this subroutine is a replacement for the CRAY subroutine second
c     for use on UNIX workstations
c     real*4 etime,x(2)
c     real*8 t
c     t=etime(x)
c     return
c     end

